####
# Compiled Glayout
# Apache License
# Version 2.0, January 2004
# http://www.apache.org/licenses/
# 2024-06-01 00:36:26.578529

from glayout.flow.pdk.mappedpdk import MappedPDK
from gdsfactory import Component
from glayout.flow.pdk.util.comp_utils import move, movex, movey, prec_ref_center, evaluate_bbox, center_to_edge_distance
from glayout.flow.pdk.util.port_utils import remove_ports_with_prefix
from glayout.flow.primitives.fet import nmos
from glayout.flow.primitives.fet import pmos
from glayout.flow.primitives.guardring import tapring
from glayout.flow.primitives.mimcap import mimcap
from glayout.flow.primitives.mimcap import mimcap_array
from glayout.flow.primitives.via_gen import via_stack
from glayout.flow.primitives.via_gen import via_array
from glayout.flow.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.flow.placement.four_transistor_interdigitized import generic_4T_interdigitzed
from glayout.flow.placement.two_transistor_interdigitized import two_pfet_interdigitized
from glayout.flow.components.diff_pair import diff_pair_generic
from glayout.flow.routing.smart_route import smart_route
from glayout.flow.routing.L_route import L_route
from glayout.flow.routing.c_route import c_route
from glayout.flow.routing.straight_route import straight_route

def CrossCoupledInverters_cell(
	pdk: MappedPDK,
	ccinvs_length: float, 
	ccinvs_fingers: int, 
):
	pdk.activate()
	CrossCoupledInverters = Component(name="CrossCoupledInverters")
	maxmetalsep = pdk.util_max_metal_seperation()
	double_maxmetalsep = 2*pdk.util_max_metal_seperation()
	triple_maxmetalsep = 3*pdk.util_max_metal_seperation()
	quadruple_maxmetalsep = 4*pdk.util_max_metal_seperation()
	# placing ccinvs centered at the origin
	ccinvs = generic_4T_interdigitzed(pdk,**{'numcols': ccinvs_fingers, 'length': ccinvs_length, 'top_row_device': "pfet", 'bottom_row_device': "nfet"})
	ccinvs_ref = prec_ref_center(ccinvs)
	CrossCoupledInverters.add(ccinvs_ref)
	CrossCoupledInverters.add_ports(ccinvs_ref.get_ports_list(),prefix="ccinvs_")
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_top_A_source_E"],CrossCoupledInverters.ports["ccinvs_top_B_source_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_bottom_A_source_E"],CrossCoupledInverters.ports["ccinvs_bottom_B_source_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_top_A_drain_E"],CrossCoupledInverters.ports["ccinvs_top_B_gate_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_bottom_A_drain_E"],CrossCoupledInverters.ports["ccinvs_bottom_B_gate_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_top_B_drain_E"],CrossCoupledInverters.ports["ccinvs_top_A_gate_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_bottom_B_drain_E"],CrossCoupledInverters.ports["ccinvs_bottom_A_gate_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_top_B_gate_E"],CrossCoupledInverters.ports["ccinvs_bottom_B_gate_E"],ccinvs_ref,CrossCoupledInverters,**{})
	CrossCoupledInverters << smart_route(pdk,CrossCoupledInverters.ports["ccinvs_top_A_gate_W"],CrossCoupledInverters.ports["ccinvs_bottom_A_gate_W"],ccinvs_ref,CrossCoupledInverters,**{})
	return CrossCoupledInverters



import warnings
from typing import Optional, Union

from gdsfactory import Component, ComponentReference
from gdsfactory.port import Port
from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.pdk.util.comp_utils import align_comp_to_port, movex
from glayout.flow.pdk.util.port_utils import (
    assert_port_manhattan,
    ports_inline,
    ports_parallel,
)
from glayout.flow.primitives.via_gen import via_stack
from glayout.flow.routing.c_route import c_route
from glayout.flow.routing.L_route import L_route
from glayout.flow.routing.straight_route import straight_route


def smart_route(
    pdk: MappedPDK,
    edge1: Port,
	edge2: Port,
    ref_comp: Optional[Union[Component, ComponentReference]]=None,
    top_comp: Optional[Union[Component, ComponentReference]]=None,
    **kwargs
) -> Component:
    # error checks
    assert_port_manhattan([edge1,edge2])
    # determine route type based on preconfiged route utils
    if top_comp is not None and ref_comp is not None:
        if ref_comp.info.get("route_genid") is not None:
            try:
                if ref_comp.info["route_genid"] == "two_transistor_interdigitized":
                    return generic_route_two_transistor_interdigitized(pdk, edge1, edge2, top_comp)
                if ref_comp.info["route_genid"] == "four_transistor_interdigitized":
                    return generic_route_four_transistor_interdigitized(pdk, edge1, edge2, top_comp)
                if ref_comp.info["route_genid"] == "common_centroid_ab_ba":
                    return generic_route_ab_ba_common_centroid(pdk, edge1, edge2, top_comp)
            except ValueError:
                warnings.warn("Attempted a specialized smart route, but failed. Now attempting general smart route...")
    # determine route type based on port orientation and distance
    if ports_parallel(edge1,edge2):
        # croute or straightroute
        if ports_inline(edge1,edge2):
            return straight_route(pdk, edge1, edge2, **kwargs)
        else:
            return c_route(pdk, edge1, edge2, **kwargs)
    else:
        return L_route(pdk, edge1, edge2, **kwargs)


# AorB_source,gate,or drain
def parse_port_name(pname: str) -> str:
    """Parse a port name to extract the device (A or B) and pin (source, drain, or gate)
    returning just that (without directions NESW)
    Args:
        pname (str): The port name to get extract device_pin
    Returns:
        str: A string containing the component and pin type separated by an underscore.
    """
    comp = str()
    pin = str()
    for part in pname.split("_"):
        if part=="A" or part=="B":
            comp = part
        if part=="source" or part=="drain" or part=="gate":
            pin=part
    return comp+"_"+pin


def check_route(name1, name2, pin1, pin2) -> bool:
    # check if this routes between the 2 pins
    cond1 = name1==pin1 and name2==pin2
    cond2 = name2==pin1 and name1==pin2
    return cond1 or cond2


def generic_route_two_transistor_interdigitized(
    pdk: MappedPDK,
    edge1: Port,
    edge2: Port,
    top_comp: Union[Component, ComponentReference]
) -> Component:
    def compensate_for_croutes(top_comp, sample_port: Port, LeftSide: bool, extend_to: Port):
        # top_comp is Component to modify, sample_port is a port name for refernce (we dont know what the prefix should be) side is either west (left) or east
        # adds a metal extension and changes the ports to compensate for the fact that there will be a croute on that side
        # extend_to should be the outside port of croute
        basename = sample_port.name.rstrip("NESW").rstrip("_")
        basename = basename.removesuffix("source").removesuffix("drain").removesuffix("gate").rstrip("_")
        basename = basename.removesuffix("A").removesuffix("B")
        direction = "_W" if LeftSide else "_E"
        #import pdb; pdb.set_trace()
        for dev in ["A_","B_"]:
            specbase = basename + dev
            for pin in ["source","drain","gate"]:
                portcorrection = top_comp.ports[specbase+pin+direction+"_private"]
                rt = top_comp << straight_route(pdk, portcorrection, extend_to, glayer2=pdk.layer_to_glayer(portcorrection.layer))
                top_comp.ports[specbase+pin+direction] = rt.ports["route"+direction]
                top_comp.ports[specbase+pin+direction].name = specbase+pin+direction
        # no return (modifies the top_comp)
    def exchange_ports(top_comp, edge: Port, direction: str) -> Port:
        # gives port which is same except a different edge N,E,S,W
        return top_comp.ports[edge.name.rstrip("NESW")+direction+"_private"]
    #glayer1 = pdk.layer_to_glayer(edge1.layer)
    glayer2 = pdk.layer_to_glayer(edge2.layer)
    name1 = parse_port_name(edge1.name)
    name2 = parse_port_name(edge2.name)
    # order names so that A is first (if only one A)
    if "A" in name2 and not("A" in name1):
        name1, name2 = name2, name1
    # trivial routes
    samecomp = any(l in name1 and l in name2 for l in ["A","B"])
    samepin = any(l in name1 and l in name2 for l in ["gate","source","drain"])
    if samecomp and samepin:
        return Component()
    # easy routes 3/15
    straight_route_width = 1 if edge1.width > 1 else edge1.width
    if check_route(name1,name2,"A_source","B_source"):
        edge = edge1 if "B" in edge1.name else edge2
        return align_comp_to_port(via_stack(pdk,"met1",glayer2),exchange_ports(top_comp,edge,"W"),alignment=("r","c"),rtr_comp_ref=False)
    if check_route(name1,name2,"A_gate","B_gate"):
        return straight_route(pdk,exchange_ports(top_comp,edge1,"S"),exchange_ports(top_comp,edge2,"S"),width=straight_route_width)
    if check_route(name1,name2,"A_drain","B_drain"):
        edge = edge1 if "A" in edge1.name else edge2
        return align_comp_to_port(via_stack(pdk,"met1",glayer2),exchange_ports(top_comp,edge,"E"),alignment=("l","c"),rtr_comp_ref=False)
    # easy self routes (A->A or B->B, and source<->drain) 2/15 (5/15)
    if check_route(name1,name2,"A_drain","A_source"):
        edge = edge1 if "drain" in edge1.name else edge2
        return align_comp_to_port(via_stack(pdk,"met1",glayer2),exchange_ports(top_comp,edge,"W"),alignment=("r","c"),rtr_comp_ref=False)
    if check_route(name1,name2,"B_drain","B_source"):
        edge = edge1 if "source" in edge1.name else edge2
        return align_comp_to_port(via_stack(pdk,"met1",glayer2),exchange_ports(top_comp,edge,"E"),alignment=("l","c"),rtr_comp_ref=False)# B is higher
    # hard self route (A->A or B->B, and source or drain -> gate) 4/15 (9/15)
    viaoffset = not(check_route(name1,name2,"A_gate","A_drain") or check_route(name1,name2,"B_gate","B_drain"))
    edge1, edge2 = (edge2, edge1) if ("drain" in edge2.name or "source" in edge2.name) else (edge1, edge2)
    width2 = edge2.width
    if check_route(name1,name2,"A_gate","A_drain") or check_route(name1,name2,"A_gate","A_source"):
        rt = c_route(pdk,exchange_ports(top_comp,edge1,"W"),exchange_ports(top_comp,edge2,"W"),viaoffset=(viaoffset,True),width2=width2) # A gets W
        compensate_for_croutes(top_comp,edge1,True,rt.ports["con_N"])
        return rt
    if check_route(name1,name2,"B_gate","B_drain") or check_route(name1,name2,"B_gate","B_source"):
        rt = c_route(pdk,exchange_ports(top_comp,edge1,"E"),exchange_ports(top_comp,edge2,"E"),viaoffset=(viaoffset,True),width2=width2) # B gets E
        compensate_for_croutes(top_comp,edge1,False,rt.ports["con_N"])
        return rt
    # inter transistor routes going to the gate of A or B   4/15 (13/15)
    if check_route(name1,name2,"A_gate","B_drain") or check_route(name1,name2,"A_gate","B_source"):
        rt = c_route(pdk,exchange_ports(top_comp,edge1,"W"),exchange_ports(top_comp,edge2,"W"),viaoffset=(not(check_route(name1,name2,"A_gate","B_drain")),True),width2=width2) # A_gate gets W
        compensate_for_croutes(top_comp,edge1,True,rt.ports["con_N"])
        return rt
    if check_route(name1,name2,"B_gate","A_drain") or check_route(name1,name2,"B_gate","A_source"):
        rt = c_route(pdk,exchange_ports(top_comp,edge1,"E"),exchange_ports(top_comp,edge2,"E"),viaoffset=(not(check_route(name1,name2,"B_gate","A_drain")),True),width2=width2) # B_gate gets E
        compensate_for_croutes(top_comp,edge1,False,rt.ports["con_N"])
        return rt
    # inter transistor routes going to from s or d to s or d   2/15 (15/15)
    if check_route(name1,name2,"A_source","B_drain"):
        edge = edge1 if "drain" in edge1.name else edge2
        return align_comp_to_port(via_stack(pdk,"met1",glayer2),exchange_ports(top_comp,edge,"W"),alignment=("r","c"),rtr_comp_ref=False)
    if check_route(name1,name2,"A_drain","B_source"):
        edge = edge1 if "drain" in edge1.name else edge2
        return align_comp_to_port(via_stack(pdk,"met1",glayer2),exchange_ports(top_comp,edge,"E"),alignment=("l","c"),rtr_comp_ref=False)
    raise ValueError("You picked a port that smart_route with interdigitized 2 transistor does not support")


def generic_route_four_transistor_interdigitized(
    pdk: MappedPDK,
    edge1: Port,
    edge2: Port,
    top_comp: Union[Component, ComponentReference]
) -> Component:
    def check_port(pname: str) -> bool:
        # check that this is a source, drain, or gate
        # returns false if the port is source drain or gate
        pname = pname.rstrip("NESW_")
        pin = pname.split("_")[-1]
        return not(pin=="source" or pin=="gate" or pin=="drain")
    def strip_portname(pname: str) -> str:
        # removes upto the two transistor interdigitzed stuff
        pname = pname.rstrip("NESW_")
        pname = pname.removesuffix("source").removesuffix("drain").removesuffix("gate")
        pname = pname.rstrip("_").removesuffix("A").removesuffix("B").rstrip("_")
        return pname
    def exchange_ports(top_comp, edge: Port, direction: str) -> Port:
        # gives port which is same except a different edge N,E,S,W
        return top_comp.ports[edge.name.rstrip("NESW")+direction+"_private"]
    def parse_port_name(pname: str) -> str:
        comp = str()
        pin = str()
        topbot = str()
        for part in pname.split("_"):
            if part=="A" or part=="B":
                comp = part
            if part=="source" or part=="drain" or part=="gate":
                pin=part
            if part=="top" or part=="bottom":
                topbot = part
        return topbot+"_"+comp+"_"+pin
    def compensate_for_croutes(top_comp, sample_port: Port, LeftSide: bool, extend_to: Port):
        # top_comp is Component to modify, sample_port is a port name for refernce (we dont know what the prefix should be) side is either west (left) or east
        # adds a metal extension and changes the ports to compensate for the fact that there will be a croute on that side
        # extend_to should be the outside port of croute
        basename = sample_port.name.rstrip("NESW").rstrip("_")
        basename = basename.removesuffix("source").removesuffix("drain").removesuffix("gate").rstrip("_")
        basename = basename.removesuffix("A").removesuffix("B")
        direction = "_W" if LeftSide else "_E"
        #import pdb; pdb.set_trace()
        for dev in ["A_","B_"]:
            specbase = basename + dev
            for pin in ["source","drain","gate"]:
                portcorrection = top_comp.ports[specbase+pin+direction+"_private"]
                rt = top_comp << straight_route(pdk, portcorrection, extend_to, glayer2=pdk.layer_to_glayer(portcorrection.layer))
                top_comp.ports[specbase+pin+direction] = rt.ports["route"+direction]
                top_comp.ports[specbase+pin+direction].name = specbase+pin+direction
    # check that this function supports the ports specified
    cond1 = check_port(edge1.name) or check_port(edge2.name)
    cond2 = strip_portname(edge1.name).split("_")[-1] != strip_portname(edge2.name).split("_")[-1]
    name1 = parse_port_name(edge1.name)
    name2 = parse_port_name(edge2.name)
    width1 = edge1.width
    width2 = edge2.width
    if "A" in name2 and not("A" in name1):
        name1, name2 = name2, name1
    if cond1:
        raise ValueError("You picked a port that smart_route with interdigitized 4 transistor does not support")
    elif cond2:
        # do your code here

        # if check_route("")

        raise ValueError("these ports will be supported soon")
        # print(f'\nname1: {name1}, name2: {name2}\n"top_A_source", "bottom_A_source"\n')
        
        print('\n result of check_route: ', check_route(name1, name2, "top_A_source", "bottom_A_source"))
        if check_route(name1, name2, "top_A_drain", "bottom_A_drain"):
            rt = c_route(pdk, exchange_ports(top_comp, edge1, "W"), exchange_ports(top_comp, edge2, "W"), viaoffset=(True, True), width1=width1, width2=width2)
            compensate_for_croutes(top_comp, edge1, True, rt.ports["con_N"])
            return rt
        if check_route(name1, name2, "top_A_drain", "bottom_A_source"):
            # print('\n\nin second if condition\n\n')
            rt = c_route(pdk, exchange_ports(top_comp, edge1, "W"), exchange_ports(top_comp, edge2, "W"), viaoffset=(True, True), width1=width1, width2=width2)
            compensate_for_croutes(top_comp, edge1, True, rt.ports["con_N"])
            return rt
        if check_route(name1, name2, "top_A_source", "bottom_A_source"):
            # print('\n\nin third if condition\n\n')
            rt = c_route(pdk, exchange_ports(top_comp, edge1, "W"), exchange_ports(top_comp, edge2, "W"), viaoffset=(True, True), width1=width1, width2=width2, extension=4*width1)
            compensate_for_croutes(top_comp, edge1, True, rt.ports["con_N"])
            return rt
        if check_route(name1, name2, "top_A_source", "bottom_A_drain"):
            # print('\n\nin fourth if condition\n\n')
            rt = c_route(pdk, exchange_ports(top_comp, edge1, "W"), exchange_ports(top_comp, edge2, "W"), viaoffset=(True, True), width1=width1, width2=width2, extension=4*width1)
            compensate_for_croutes(top_comp, edge1, True, rt.ports["con_N"])
            return rt
    else:
        # else return 2 tran route
        return generic_route_two_transistor_interdigitized(pdk, edge1, edge2, top_comp)



def generic_route_ab_ba_common_centroid(
    pdk: MappedPDK,
    edge1: Port,
    edge2: Port,
    top_comp: Union[Component, ComponentReference]
) -> Component:
    # TODO: implement
    name1, name2 = parse_port_name(edge1.name), parse_port_name(edge2.name)
    width1 = edge1.width
    # order names so that A is first (if only one A)
    if "A" in name2 and not("A" in name1):
        name1, name2 = name2, name1
    # same device routes (A->A or B->B) (6/15)
    if check_route(name1,name2,"A_source","A_gate"):
        return straight_route(pdk, top_comp.ports["A_source_E_private"],top_comp.ports["A_gate_route_con_N"],via2_alignment=("right","bottom"))
    if check_route(name1,name2,"A_drain","A_gate"):
        return straight_route(pdk, top_comp.ports["A_drain_E_private"],top_comp.ports["A_gate_route_con_N"],via2_alignment=("right","top"))
    if check_route(name1,name2,"A_source","A_drain"):
        straight_route(pdk, top_comp.ports["br_multiplier_0_source_N"],top_comp.ports["br_multiplier_0_drain_S"],width=min(width1,1))
        return straight_route(pdk, top_comp.ports["tl_multiplier_0_source_S"],top_comp.ports["tl_multiplier_0_drain_N"],width=min(width1,1))
    if check_route(name1,name2,"B_source","B_gate"):
        return straight_route(pdk, top_comp.ports["B_source_W_private"],top_comp.ports["B_gate_route_con_N"],via2_alignment=("left","bottom"))
    if check_route(name1,name2,"B_drain","B_gate"):
        return straight_route(pdk, top_comp.ports["B_drain_W_private"],top_comp.ports["B_gate_route_con_N"],via2_alignment=("left","top"))
    if check_route(name1,name2,"B_source","B_drain"):
        top_comp << straight_route(pdk, top_comp.ports["tr_multiplier_0_source_S"],top_comp.ports["tr_multiplier_0_drain_N"],width=min(width1,1))
        return straight_route(pdk, top_comp.ports["bl_multiplier_0_source_N"],top_comp.ports["bl_multiplier_0_drain_S"],width=min(width1,1))
    # A_src/drain->B_gate or B_src/drain->A_gate (4/15)
    if check_route(name1,name2,"A_source","B_gate"):
        return straight_route(pdk, top_comp.ports["A_source_W_private"],top_comp.ports["B_gate_route_con_N"],via2_alignment=("left","top"))
    if check_route(name1,name2,"A_drain","B_gate"):
        return straight_route(pdk, top_comp.ports["A_drain_W_private"],top_comp.ports["B_gate_route_con_N"],via2_alignment=("left","bottom"))
    if check_route(name1,name2,"B_source","A_gate"):
        return straight_route(pdk, top_comp.ports["B_source_E_private"],top_comp.ports["A_gate_route_con_N"],via2_alignment=("right","top"))
    if check_route(name1,name2,"B_drain","A_gate"):
        return straight_route(pdk, top_comp.ports["B_drain_E_private"],top_comp.ports["A_gate_route_con_N"],via2_alignment=("right","bottom"))
    # A_src/drain->B_src or A_src/drain->B_drain (4/15)
    if check_route(name1,name2,"A_source","B_source"):
        return straight_route(pdk, top_comp.ports["tl_multiplier_0_source_E"],top_comp.ports["tr_multiplier_0_source_W"])
    if check_route(name1,name2,"A_drain","B_source"):
        portmv1 = top_comp.ports["tl_multiplier_0_drain_E"].copy()
        return straight_route(pdk, top_comp.ports["tl_multiplier_0_drain_E"],movex(portmv1,2*pdk.get_grule(pdk.layer_to_glayer(portmv1.layer))["min_separation"]))
    if check_route(name1,name2,"A_source","B_drain"):
        portmv1 = top_comp.ports["tr_multiplier_0_drain_W"].copy()
        return straight_route(pdk, top_comp.ports["tr_multiplier_0_drain_W"],movex(portmv1,-2*pdk.get_grule(pdk.layer_to_glayer(portmv1.layer))["min_separation"]))
    if check_route(name1,name2,"A_drain","B_drain"):
        portmv1 = top_comp.ports["bl_mutliplier_0_drain_N"].copy()
        portmv2 = top_comp.ports["br_multiplier_0_drain_N"].copy()
        top_comp << straight_route(pdk, movex(portmv1,-portmv1.width/2), top_comp.ports["tl_multiplier_0_drain_S"],width=width1)
        return straight_route(pdk, movex(portmv2,portmv2.width/2),top_comp.ports["tr_multiplier_0_drain_S"])
    # A_gate -> B_gate (1/15)
    if check_route(name1,name2,"A_gate","B_gate"):
        return straight_route(pdk,top_comp.ports["br_multiplier_0_gate_W"],top_comp.ports["bl_multiplier_0_gate_E"])
    raise ValueError("You picked a port that smart_route with ab_ba_common_centroid does not support")



# this file contains functions which dynamically create and import cells, and run sessions
import importlib.util
import inspect
import os
import re
from typing import Callable, Union
from pathlib import Path
import tempfile
import glayout.syntaxer.relational

from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.pdk.util.port_utils import PortTree
import sys
import traceback
import glayout.syntaxer.process_input


def reload_saved_convo(convo_file: Union[Path, str]) -> tuple:
    """restores a conversation from a .convo file

    Args:
        convo_file (Union[Path, str]): path to the .convo file

    Returns tuple:
        Session: restored conversation object
        loop_count: prompt index after restoring the conversation
    """
    savedconvo = Path(convo_file).resolve()
    if not savedconvo.is_file():
        raise FileNotFoundError("load conversation should be from an existing file")
    with savedconvo.open("r") as loadconvo:
        lines = loadconvo.readlines()
        for i, line in enumerate(lines):
            if i == 0:
                convo = glayout.syntaxer.process_input.Session(inputstream=sys.stdin, outputstream=sys.stdout, toplvlname=line)
                continue
            convo.process_next_input(line)
        loop_count = len(lines) - 1
    print("\n\nloaded conversation from " + str(savedconvo))
    return convo, loop_count


def run_session(load_conversation: Union[str, Path], restore_and_exit: bool=False) -> str:
    """Manage, interact, and run conversation sessions from command line

    Returns:
        str: Glayout python code corresponding to the current session
    """
    # start convo and intialize loop counter (will be updated if we read from file)
    # if saved convo then load everything from save file, else read form stdio
    if load_conversation is not None:
        convo, loop_count = reload_saved_convo(load_conversation)
        if restore_and_exit:
            return convo.code.get_code()
    else:
        convo, loop_count = glayout.syntaxer.process_input.Session(inputstream=sys.stdin, outputstream=sys.stdout), int(0)
    # enter design loop
    session_ongoing = True
    while session_ongoing:
        convo.print_to_stream("\ntask " + str(loop_count))
        loop_count += 1
        convo.print_to_stream(convo.generic_prompt)
        try:
            session_ongoing = convo.process_next_input(convo.read_from_stream())
        except Exception as e:
            print(traceback.format_exc())
            print("an exception was encounterd")
            print(str(e))
            print("restoring last valid state and resuming regular program execution\n")
            convo = convo.backup
            loop_count -= 1
            session_ongoing = True
    return convo.code.get_code()



def get_default_arguments(func: Callable, pdk: MappedPDK) -> dict:
    """Gets default arguments to a function based on its argument types.

    Args:
        func (callable): The function to which default arguments will be added.
        pdk (MappedPDK): If one of the non default args is of type MappedPDK, then this pdk is used for default value

    Returns:
        dict: A dictionary containing default arguments with values based on the argument types of the input function.
    """
    # get args that dont have a default
    argspec = inspect.getfullargspec(func)
    args_with_defaults = argspec.defaults or []
    num_args_without_defaults = len(argspec.args) - len(args_with_defaults)
    args_without_defaults = argspec.args[:num_args_without_defaults]
    # loop through non default args and try to set some value for them
    kwargs = {}
    for arg, arg_type in zip(args_without_defaults, argspec.annotations.values()):
        # pick some default value
        if arg_type == int:
            default_value = 2
        elif arg_type == float:
            default_value = 2.5
        elif arg_type == bool:
            default_value = True
        elif arg_type == str:
            default_value = "met1"
        # hard to guess what a person might want with a tuple, but make it a pair of ints (size) is a safe bet
        elif arg_type == tuple:
            default_value = (1,1)
        # hard to guess what a person might want with a list, but make it a long list of ints is a safe bet
        elif arg_type == list:
            default_value = [1,1,1,1,1,1,1,1,1]
        elif arg_type == MappedPDK:
            default_value = pdk
        else:  # for other types, set default to None
            default_value = None
        # add this argument to the kwargs
        kwargs[arg] = default_value
    return kwargs


def get_funccell_name(glayout_code: str) -> str:
    pattern = r"def\s+([a-zA-Z_][a-zA-Z0-9_]*)cell\s*\("
    return re.search(pattern, glayout_code).group().lstrip("def").rstrip("(").strip()

class CodeImportHandler:
    """create, manage, destroy temporary files created as part of dynamic importing
    contains
        self.function (Callable): the function handle
        self.func_name (str): the name of the function imported
        self.temp_module: the imported module handle
    """

    def __init__(self, glayout_code: Union[str, Path]):
        """create temporary file with glayout python code from glayout_code string
        and import the module
        Args:
            glayout_code (str, Path): string containing cell function and imports.
            ****or, string or pathlib.Path: path to a convo file
        """
        # check if this is a code string or a file
        precompiled_code = False
        if isinstance(glayout_code, str):
            if glayout.syntaxer.relational.GlayoutCode.HEAD_MARKER in glayout_code:
                precompiled_code = True
        # if this is not a code string, convert NLP file to python code
        if not precompiled_code:
            glayout_code = run_session(glayout_code, restore_and_exit=True)
        # figure out what the cell is called
        self.func_name = get_funccell_name(glayout_code)
        pymodule_name = self.func_name.removesuffix("_cell") + ".py"
        # create the cell py module
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir_path = Path(temp_dir).resolve()
            pythonfile = temp_dir_path / pymodule_name
            with pythonfile.open(mode="w") as pyfile:
                pyfile.write(glayout_code)
            # import the cell
            spec = importlib.util.spec_from_file_location(self.func_name, pythonfile)
            self.temp_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(self.temp_module)
            self.function = getattr(self.temp_module, self.func_name)



def run_glayout_code_cell(pdk: MappedPDK, glayout_code: str) -> bool:
    """Instantiate a layout from the given glayout cell code and returns Component

    Args:
        pdk (MappedPDK): pdk to instantiate this cell
        glayout_code (str): string containg the cell function and all needed imports

    Returns:
        Component: gdsfactory Component corresponding to the produced layout
    """
    testcell = CodeImportHandler(glayout_code).function
    return testcell(**get_default_arguments(testcell, pdk))


def show_glayout_code_cell(pdk: MappedPDK, glayout_code: str):
    """Instantiate and show a layout from the given glayout cell code

    Args:
        pdk (MappedPDK): pdk to instantiate this cell
        glayout_code (str): string containg the cell function and all needed imports
    """
    run_glayout_code_cell(pdk, glayout_code).show()


def getPortTree_glayout_code_cell(pdk: MappedPDK, glayout_code: str) -> PortTree:
    """return PortTree for a given glayout cell

    Args:
        pdk (MappedPDK): pdk to instantiate this cell
        glayout_code (str): string containg the cell function and all needed imports
    
    Returns PortTree
    """
    return PortTree(run_glayout_code_cell(pdk, glayout_code))

def printPortTree_glayout_code_cell(pdk: MappedPDK, glayout_code: str):
    """return PortTree for a given glayout cell

    Args:
        pdk (MappedPDK): pdk to instantiate this cell
        glayout_code (str): string containg the cell function and all needed imports
    """
    PortTree(run_glayout_code_cell(pdk, glayout_code)).print(depth=6)



# 4 transistor placed in two rows (each row is an interdigitized pair of transistors)
# the 4 transistors are labeled top or bottom and transistor A or B
# top_A_, bottom_A, top_B_, bottom_B_

from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.placement.two_transistor_interdigitized import two_nfet_interdigitized, two_pfet_interdigitized
from typing import Literal, Optional
from gdsfactory import Component
from glayout.flow.pdk.util.comp_utils import evaluate_bbox, movey
from glayout.flow.primitives.guardring import tapring

def generic_4T_interdigitzed(
    pdk: MappedPDK,
    top_row_device: Literal["nfet", "pfet"],
    bottom_row_device: Literal["nfet", "pfet"],
    numcols: int,
    length: float=None,
    with_substrate_tap: bool = True,
    top_kwargs: Optional[dict]=None,
    bottom_kwargs: Optional[dict]=None
):
    if top_kwargs is None:
        top_kwargs = dict()
    if bottom_kwargs is None:
        bottom_kwargs = dict()
    # place
    toplvl = Component()
    if top_row_device=="nfet":
        toprow = toplvl << two_nfet_interdigitized(pdk,numcols,with_substrate_tap=False,length=length,**top_kwargs)
    else:
        toprow = toplvl << two_pfet_interdigitized(pdk,numcols,with_substrate_tap=False,length=length,**top_kwargs)
    if bottom_row_device=="nfet":
        bottomrow = toplvl << two_nfet_interdigitized(pdk,numcols,with_substrate_tap=False,length=length,**bottom_kwargs)
    else:
        bottomrow = toplvl << two_pfet_interdigitized(pdk,numcols,with_substrate_tap=False,length=length,**bottom_kwargs)
    # move
    toprow.movey(pdk.snap_to_2xgrid((evaluate_bbox(bottomrow)[1]/2 + evaluate_bbox(toprow)[1]/2 + pdk.util_max_metal_seperation())))
    # add substrate tap
    if with_substrate_tap:
        substrate_tap = tapring(pdk, enclosed_rectangle=pdk.snap_to_2xgrid(evaluate_bbox(toplvl.flatten(),padding=pdk.util_max_metal_seperation())))
        substrate_tap_ref = toplvl << movey(substrate_tap,destination=pdk.snap_to_2xgrid(toplvl.flatten().center[1],snap4=True))
    # add ports
    toplvl.add_ports(substrate_tap_ref.get_ports_list(),prefix="substratetap_")
    toplvl.add_ports(toprow.get_ports_list(),prefix="top_")
    toplvl.add_ports(bottomrow.get_ports_list(),prefix="bottom_")
    # flag for smart route
    toplvl.info["route_genid"] = "four_transistor_interdigitized"
    return toplvl



"""
usage: from mappedpdk import MappedPDK
"""

from gdsfactory.pdk import Pdk
from gdsfactory.typings import Component, PathType, Layer
from pydantic import validator, StrictStr, ValidationError
from typing import ClassVar, Optional, Any, Union, Literal, Iterable, TypedDict
from pathlib import Path
from decimal import Decimal, ROUND_UP
import tempfile
import subprocess
from decimal import Decimal
from pydantic import validate_arguments
import xml.etree.ElementTree as ET
import pathlib, shutil, os, sys

class SetupPDKFiles:
    """Class to setup the PDK files required for DRC and LVS checks.
    """

    def __init__(
        self, 
        pdk_root: Optional[PathType] = None, 
        klayout_drc_file: Optional[PathType] = None, 
        lvs_schematic_ref_file: Optional[PathType] = None,
        lvs_setup_tcl_file: Optional[PathType] = None, 
        magic_drc_file: Optional[PathType] = None,
        temp_dir: Optional[PathType] = None,
        pdk: Optional[str] = 'sky130'
    ):
        """Initializes the class with the required PDK files for DRC and LVS checks."""
        self.pdk = pdk
        self.temp_dir = temp_dir
        
        if pdk_root is None: 
            # condition for pdk_root not being provided
            if klayout_drc_file is None: 
                raise ValueError("Please provide a Klayout DRC file if not providing PDK root!")
            else:
                if not isinstance(klayout_drc_file, PathType):
                    raise ValueError("The file must be provided as a Path object or string")
                else:
                    self.klayout_drc_file = klayout_drc_file
            if magic_drc_file is None:
                raise ValueError("Please provide a magic DRC file")
            else:
                if not isinstance(magic_drc_file, PathType):
                    raise ValueError("The files must be provided as Path objects or strings")
                else:
                    self.magic_drc_file = magic_drc_file
            if lvs_schematic_ref_file is None or lvs_setup_tcl_file is None or magic_drc_file is None:
                raise ValueError(f"""Please provide the following files:
                      - a spice file with subckt references(lvs_schematic_ref_file)
                      - a tcl file with setup commands(lvs_setup_tcl_file)
                      - a magic DRC file(magic_drc_file)""")
            else:
                if not isinstance(lvs_schematic_ref_file, PathType) or not isinstance(lvs_setup_tcl_file, PathType):
                    raise ValueError("The files must be provided as Path objects or strings")
                else:
                    self.lvs_schematic_ref_file = lvs_schematic_ref_file
                    self.lvs_setup_tcl_file = lvs_setup_tcl_file
                    self.magic_drc_file = magic_drc_file
        else:
            if not isinstance(pdk_root, PathType):
            # condition for pdk_root not being a Path object
                raise ValueError("pdk_root must be a Path object or string")
            # condition for pdk_root being provided
            self.pdk_root = pdk_root
            if klayout_drc_file is None:
                # condition for klayout_drc_file not being provided with pdk_root provided
                if pdk == 'sky130':
                    klayout_drc_file = Path(pdk_root) / "sky130A" / "libs.tech" / "klayout" / "drc" / "sky130.lydrc"
                elif pdk == 'gf180':
                    raise NotImplementedError("Klayout lvs is not supported for gf180 PDK")
                else:
                    raise ValueError("pdk must be either 'sky130' or 'gf180', others not supported!")
            
                
            if lvs_schematic_ref_file is None:
                if pdk == 'gf180':
                    raise NotImplementedError("LVS is not supported for gf180 PDK")
                
                lvs_spice_file = Path(pdk_root) / "sky130A" / "libs.ref" / "sky130_fd_sc_hd" / "spice" / "sky130_fd_sc_hd.spice"
                lvs_schematic_ref_file = temp_dir / "sky130_fd_sc_hd.spice"
                self.write_custom_spice_to_file(lvs_spice_file, lvs_schematic_ref_file)
            
            if lvs_setup_tcl_file is None:
                if pdk == 'gf180':
                    raise NotImplementedError("LVS is not supported for gf180 PDK")
                dest_lvs_setup_tcl = temp_dir / "sky130A_setup.tcl"
                lvs_setup_tcl_file = self.magic_netgen_file_exists(dest_lvs_setup_tcl, pdk_root)
                
            if magic_drc_file is None:
                if pdk == "sky130":
                    dest_magic_drc = temp_dir / f"{pdk}A.magicrc"     
                elif pdk == "gf180": 
                    dest_magic_drc = temp_dir / f"{pdk}mcuC.magicrc"
                    
                magic_drc_file = self.magic_netgen_file_exists(dest_magic_drc, pdk_root)
                
            self.klayout_drc_file = klayout_drc_file
            self.lvs_schematic_ref_file = lvs_schematic_ref_file
            self.lvs_setup_tcl_file = lvs_setup_tcl_file
            self.magic_drc_file = magic_drc_file
            
    def write_custom_spice_to_file(
        self, 
        input_spice: str, 
        output_spice: str
    ):
        """Writes a custom spice file to the output_spice file path."""
        custom_spice_string = f"""
* HEADER and SLC cells
* Proxy pins that may or may not be generated by magic sky130_fd_sc_hd__tap_1_0/VPB sky130_fd_sc_hd__tap_1_1/VPB

.subckt HEADER VGND VIN VNB VPWR
xMN0 VPWR VGND net7 VNB sky130_fd_pr__nfet_03v3_nvt ad=0p pd=0u as=0p ps=0u w=700000u l=500000u
xMN1 net7 VGND VPWR VNB sky130_fd_pr__nfet_03v3_nvt ad=1.1445e+12p pd=1.167e+07u as=3.92e+11p ps=3.92e+06u w=700000u l=500000u
xMN2 VPWR VGND net7 VNB sky130_fd_pr__nfet_03v3_nvt ad=0p pd=0u as=0p ps=0u w=700000u l=500000u
xMN3 net7 VGND VPWR VNB sky130_fd_pr__nfet_03v3_nvt ad=0p pd=0u as=0p ps=0u w=700000u l=500000u
xMN4 net7 VGND VIN VNB sky130_fd_pr__nfet_03v3_nvt ad=0p pd=0u as=0p ps=0u w=700000u l=500000u
xMN5 VIN VGND net7 VNB sky130_fd_pr__nfet_03v3_nvt ad=3.92e+11p pd=3.92e+06u as=0p ps=0u w=700000u l=500000u
xMN6 net7 VGND VIN VNB sky130_fd_pr__nfet_03v3_nvt ad=0p pd=0u as=0p ps=0u w=700000u l=500000u
xMN7 VIN VGND net7 VNB sky130_fd_pr__nfet_03v3_nvt ad=0p pd=0u as=0p ps=0u w=700000u l=500000u
.ends

.subckt SLC IN INB VOUT VGND VNB VPWR VPB
xMP0 net02 net07 VPWR VPB sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMP1 net03 net03 net02 VPB sky130_fd_pr__pfet_01v8_hvt mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMP3 net06 net03 VPWR VPB sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMP4 net07 net07 net06 VPB sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMP6 VOUT net06 VPWR VPB sky130_fd_pr__pfet_01v8_hvt m=2 mult=1 w=0.5 l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMN4 VOUT net07 VGND VNB sky130_fd_pr__nfet_01v8 m=1 mult=1 w=0.65 l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMN0 net03 INB VGND VNB sky130_fd_pr__nfet_01v8_lvt m=10 mult=1 w=0.5 l=150000u sa=265e-3 sb=265e-3 sd=280e-3
xMN8 net07 IN VGND VNB sky130_fd_pr__nfet_01v8_lvt m=10 mult=1 w=500000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
.ends

******* EOF

* Modified subthreshold inverter cell

.subckt dinv1 Yb Y Ab A Apb Ap VGND VPWR
X1    t1    Apb    VPWR    VPWR   sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
X2    Y    Yb    t1    t1   sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
X3    t2    Ap    VPWR    VPWR  sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3 
X4    Yb    Y    t2    t2   sky130_fd_pr__pfet_01v8_hvt m=1 mult=1 w=360000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
X5    Y    Ab    t3    t3   sky130_fd_pr__nfet_01v8 m=1 mult=1 w=180000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
X6    t3    Ab    VGND   VGND  sky130_fd_pr__nfet_01v8 m=1 mult=1 w=180000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
X7    Yb    A    t4    t4  sky130_fd_pr__nfet_01v8 m=1 mult=1 w=180000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
X8    t4    A    VGND    VGND  sky130_fd_pr__nfet_01v8 m=1 mult=1 w=180000u l=150000u sa=265e-3 sb=265e-3 sd=280e-3
.ends 

******** EOF
"""
        with open(input_spice, 'r') as rf:
            lines = rf.readlines()
            with open(output_spice, 'w') as wf:
                for line in lines:
                    wf.write(line)
                wf.write(custom_spice_string)
                
    def magic_netgen_file_exists(
        self, 
        magic_file, 
        pdk_root
    ):
            """Check that magic_file exists if not none, and copy it to the required pdk directory if it doesn't exist"""
            def copy_if_not_exists(src, dest):
                if not os.path.exists(dest):
                    print(f'copying file from required pdk dir: {src}')
                    shutil.copy(src, dest)
                else:
                    print(f'Either file not found: {src}, or already copied to: {dest}')
                    
            def return_pdk_full_name(self):
                """Returns the full name of the pdk based on the class name"""
                if self.pdk == 'sky130':
                    return f"{self.pdk}A"
                elif self.pdk == 'gf180':
                    return f"{self.pdk}mcuC"
            
            # Check if the magic_file exists, if not, copy it to the required pdk directory
            if magic_file != None and not magic_file.is_file():
                # Check if the pdk_root is a directory, if not raise an error
                if pdk_root != None and Path(pdk_root).resolve().is_dir():
                    pdk_name = return_pdk_full_name(self)
                    if ".magicrc" in str(magic_file):
                        magic_file_source = Path(pdk_root) / f"{pdk_name}" / "libs.tech" / "magic" / f"{pdk_name}.magicrc"
                    elif "_setup.tcl" in str(magic_file):
                        magic_file_source = Path(pdk_root) / f"{pdk_name}" / "libs.tech" / "netgen" / f"{pdk_name}_setup.tcl"
                    copy_if_not_exists(magic_file_source, magic_file)
                    return magic_file
                elif pdk_root != None and not pdk_root.is_dir():
                    raise ValueError("pdk_root must be a directory")
                raise ValueError("magic/netgen script: the path given is not a file")
            return magic_file
        
    def return_dict_of_files(
        self
    ):
        pdk_files = {
            'pdk_root': self.pdk_root,
            'klayout_drc_file': self.klayout_drc_file,
            'lvs_schematic_ref_file': self.lvs_schematic_ref_file,
            'lvs_setup_tcl_file': self.lvs_setup_tcl_file,
            'magic_drc_file': self.magic_drc_file,
            'temp_dir': self.temp_dir,
            'pdk': self.pdk
        }
        return pdk_files
        

class MappedPDK(Pdk):
    """Inherits everything from the pdk class but also requires mapping to glayers
    glayers are generic layers which can be returned with get_glayer(name: str)
    has_required_glayers(list[str]) is used to verify all required generic layers are
    present"""

    valid_glayers: ClassVar[tuple[str]] = (
        "dnwell",
        "pwell",
        "nwell",
        "p+s/d",
        "n+s/d",
        "active_diff",
        "active_tap",
        "poly",
        "mcon",
        "met1",
        "via1",
        "met2",
        "via2",
        "met3",
        "via3",
        "met4",
        "via4",
        "met5",
        "capmet",
    )

    models: dict = {
        "nfet": "",
        "pfet": "",
        "mimcap": ""
    }

    glayers: dict[StrictStr, Union[StrictStr, tuple[int,int]]]
    # friendly way to implement a graph
    grules: dict[StrictStr, dict[StrictStr, Optional[dict[StrictStr, Any]]]]
    pdk_files: dict[StrictStr, Union[PathType, None]]

    @validator("models")
    def models_check(cls, models_obj: dict[StrictStr, StrictStr]):
        for model in models_obj.keys():
            if not model in ["nfet","pfet","mimcap"]:
                raise ValueError(f"specify nfet, pfet, or mimcap models only")
        return models_obj

    @validator("glayers")
    def glayers_check_keys(cls, glayers_obj: dict[StrictStr, Union[StrictStr, tuple[int,int]]]):
        """force people to pick glayers from a finite set of string layers that you define
        checks glayers to ensure valid keys,type. Glayers must be passed as dict[str,str]
        if someone tries to pass a glayers dict that has a bad key, throw an error"""
        for glayer, mapped_layer in glayers_obj.items():
            if (not isinstance(glayer, str)) or (not isinstance(mapped_layer, Union[str, tuple])):
                raise TypeError("glayers should be passed as dict[str, Union[StrictStr, tuple[int,int]]]")
            if glayer not in cls.valid_glayers:
                raise ValueError(
                    "glayers keys must be one of generic layers listed in class variable valid_glayers"
                )
        return glayers_obj

    @validate_arguments
    def drc(
        self,
        layout: Component | PathType,
        output_dir_or_file: Optional[PathType] = None,
    ):
        """Returns true if the layout is DRC clean and false if not
        Also saves detailed results to output_dir_or_file location as lyrdb
        layout can be passed as a file path or gdsfactory component"""
        if not self.klayout_lydrc_file:
            raise NotImplementedError("no drc script for this pdk")
        # find layout gds file path
        tempdir = None
        if isinstance(layout, Component):
            tempdir = tempfile.TemporaryDirectory()
            layout_path = Path(layout.write_gds(gdsdir=tempdir.name)).resolve()
        elif isinstance(layout, PathType):
            layout_path = Path(layout).resolve()
        else:
            raise TypeError("layout should be a Component, Path, or string")
        if not layout_path.is_file():
            raise ValueError("layout must exist, the path given is not a file")
        # find report file path, if None then use current directory
        report_path = (
            Path(output_dir_or_file).resolve()
            if output_dir_or_file
            else Path.cwd().resolve()
        )
        if report_path.is_dir():
            report_path = Path(
                report_path
                / str(
                    self.name
                    + layout_path.name.replace(layout_path.suffix, "")
                    + "_drcreport.lyrdb"
                )
            )
        elif not report_path.is_file():
            raise ValueError("report_path must be file or dir")
        # run klayout drc
        drc_args = [
            "klayout",
            "-b",
            "-r",
            str(self.pdk_files['klayout_lydrc_file']),
            "-rd",
            "input=" + str(layout_path),
            "-rd",
            "report=" + str(report_path),
        ]
        rtr_code = subprocess.Popen(drc_args).wait()
        if rtr_code:
            raise RuntimeError("error running klayout DRC")
        # clean up and return
        if tempdir:
            tempdir.cleanup()
        # there is a drc parsing open-source at:
        # https://github.com/google/globalfoundries-pdk-libs-gf180mcu_fd_pr/blob/main/rules/klayout/drc
        # eventually I can return more info on the drc run, but for now just void and view the lyrdb in klayout

        # Open DRC output XML file
        drc_tree = ET.parse(report_path.resolve())
        drc_root = drc_tree.getroot()
        if drc_root.tag != "report-database":
            raise TypeError("DRC report file is not a valid report-database")
        # Check if DRC passed
        drc_error_count = len(drc_root[7])
        return (drc_error_count == 0)

    @validate_arguments
    def drc_magic(
        self, 
        layout: Component | str, 
        design_name: str, 
        pdk_root: Optional[PathType] = None, 
        magic_drc_file: Optional[PathType] = None
    ):
        """Runs DRC using magic on the either the component or the gds file path provided. Requires the design name and the pdk_root to be specified, handles importing the required magicrc and other setup files, if not specified. Accepts overriden magic_commands_file and magic_drc_file.

        Args:
            - layout (Component | str):
                - Either the Component or the gds file path to run DRC on
            - design_name (str):
                - The designated name of the design
            - pdk_root (Optional[PathType], optional):
                - The directory where the pdk files are located. 
                - e.g. - "/usr/bin/miniconda3/share/pdk/". Defaults to "/usr/bin/miniconda3/share/pdk/".
            - magic_drc_file (Optional[PathType], optional):
                - The .magicrc file for your PDK of choice. 
                - Defaults to None.

        Raises:
            - ValueError: 
                - If the magic DRC cannot be run! 
                - Please either provide a PDK root or the following files: 
                    - a file containing magic commands to be executed for DRC (magic_commands.tcl) 
                    - the .magicrc file for your PDK of choice
        """
        
        def create_magic_commands_file(temp_dir):
            # magic commands file creation
            print("Defaulting to stale magic_commands.tcl")
            magic_commands_file_str = f"""
gds flatglob *$$*
gds flatglob *VIA*
gds flatglob *CDNS*
gds flatglob *capacitor_test_nf*

gds read $::env(RESULTS_DIR)/$::env(DESIGN_NAME).gds

proc custom_drc_save_report {{{{cellname ""}} {{outfile ""}}}} {{

if {{$outfile == ""}} {{set outfile "drc.out"}}

set fout [open $outfile w]
set oscale [cif scale out]
if {{$cellname == ""}} {{
    select top cell
    set cellname [cellname list self]
    set origname ""
}} else {{
    set origname [cellname list self]
    puts stdout "\[INFO\]: Loading $cellname\n"
    flush stdout

    load $cellname
    select top cell
}}

drc check
set count [drc list count]

puts $fout "$cellname count: $count"
puts $fout "----------------------------------------"
set drcresult [drc listall why]
foreach {{errtype coordlist}} $drcresult {{
    puts $fout $errtype
    puts $fout "----------------------------------------"
    foreach coord $coordlist {{
        set bllx [expr {{$oscale * [lindex $coord 0]}}]
        set blly [expr {{$oscale * [lindex $coord 1]}}]
        set burx [expr {{$oscale * [lindex $coord 2]}}]
        set bury [expr {{$oscale * [lindex $coord 3]}}]
        set coords [format " %.3fum %.3fum %.3fum %.3fum" $bllx $blly $burx $bury]
        puts $fout "$coords"
    }}
puts $fout "----------------------------------------"
}}
puts $fout ""

if {{$origname != ""}} {{
    load $origname
}}

close $fout
puts "\[INFO\]: DONE with $outfile\n"
}}

custom_drc_save_report $::env(DESIGN_NAME) $::env(REPORTS_DIR)/$::env(DESIGN_NAME).rpt
"""
                
            new_path = temp_dir / "magic_commands.tcl"
            with open(str(new_path.resolve()), 'w') as f:
                f.write(magic_commands_file_str)
                
            return new_path.resolve()
    
        self.pdk_files['pdk'] = self.name
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir_path = pathlib.Path(temp_dir).resolve()
            self.pdk_files['temp_dir'] = temp_dir_path
            
            if pdk_root is None:
                print("using default pdk_root: /usr/bin/miniconda3/share/pdk/")
            else: 
                print('using provided pdk_root')
                self.pdk_files['pdk_root'] = pdk_root
                
            env_vars = {
                    'PDK_ROOT': str(self.pdk_files['pdk_root']),
                    'DESIGN_NAME': design_name,
                    'REPORTS_DIR': str(temp_dir_path),
                    'RESULTS_DIR': str(temp_dir_path)
                }
            os.environ.update(env_vars)
                    
            gds_path = str(temp_dir_path / f"{design_name}.gds")
            if isinstance(layout, Component):
                layout.write_gds(gds_path)
            elif isinstance(layout, PathType):            
                shutil.copy(layout, gds_path)
            
            magicrc_file = self.pdk_files['magic_drc_file'] if magic_drc_file is None else magic_drc_file
            magic_cmd_file = create_magic_commands_file(temp_dir_path)
            cmd = f'bash -c "magic -rcfile {magicrc_file} -noconsole -dnull {magic_cmd_file} < /dev/null"'
            
            subp = subprocess.Popen(
                cmd, 
                shell=True, 
                stdout=subprocess.PIPE, 
                stderr=subprocess.PIPE
            )
            
            subp.wait()
            print(subp.stdout.read().decode('utf-8'))
            
            subproc_code = subp.returncode
            result_str = "magic drc script passed" if subproc_code == 0 else "magic drc script failed"
            # print errors
            
            errors = subp.stderr.read().decode('utf-8')
            if errors:
                print(f"Soft errors: \n{errors}")
                
            with open(f'{str(temp_dir_path)}/{design_name}.rpt', 'r') as f:
                num_lines = len(f.readlines())
                if num_lines > 3:
                    result_str = result_str + "\nErrors found in DRC report"
                    f.seek(0)
                    print(f.read())
                else:
                    result_str = result_str + "\nNo errors found in DRC report"

        print(result_str)

    @validate_arguments
    def lvs_netgen(
        self,
        layout: Component | PathType, 
        design_name: str, 
        pdk_root: Optional[PathType] = None,
        lvs_setup_tcl_file: Optional[PathType] = None,
        lvs_schematic_ref_file: Optional[PathType] = None,
        magic_drc_file: Optional[PathType] = None, 
        netlist: Optional[PathType] = None,
        report_handling: Optional[tuple[Optional[bool], Optional[str]]] = (False, None)
    ):
        """Runs LVS using netgen on the either the component or the gds file path provided. Requires the design name and the pdk_root to be specified, handles importing the required setup files, if not specified. Accepts overriden lvs_setup_tcl_file, lvs_schematic_ref_file, and magic_drc_file.

        Args:
            - layout (Component | PathType): 
                - Either the Component or the gds file path to run LVS on
            design_name (str): 
                - The designated name of the design
            pdk_root (Optional[PathType], optional): 
                - The directory where the pdk files are located. 
                - e.g. - "/usr/bin/miniconda3/share/pdk/". Defaults to None.
            - lvs_setup_tcl_file (Optional[PathType], optional): 
                - A file containing netgen setup commands to be executed for LVS (sky130A_setup.tcl). 
                - Defaults to None. 
                - (Found in libs.tech/netgen/ of the pdk directory)
            - lvs_schematic_ref_file (Optional[PathType], optional): 
                - A spice file containing subckt definitions for the PDK. 
                - Defaults to None. 
                - (Found in libs.ref/sky130_fd_sc_hd/spice of the pdk directory)
            - magic_drc_file (Optional[PathType], optional): 
                - The .magicrc file for your PDK of choice. 
                - Defaults to None.
            - netlist (Optional[PathType], optional): 
                - The path to the CDL file (netlist) for the design. 
                - Defaults to None.
            - report_handling (Optional[tuple[Optional[bool], Optional[str]]], optional): 
                - A tuple containing a boolean value and a string. 
                - The boolean value indicates whether to handle the report or not, and the string indicates the path to the report. 
                - Defaults to (False, None).

        Raises:
            - NotImplementedError: 
                - If LVS is not supported for gf180 PDK
            - RuntimeError:
                - If Netgen or Magic is not found in the system
            - ValueError:
                - If LVS cannot be run! 
                - Please provide either a PDK root or the following files: a setup.tcl file for the PDK, a .magicrc file, and a spice file containing subckt definitions for the PDK.

        """
        if not self.name == 'sky130':
            raise NotImplementedError("LVS only supported for sky130 PDK")
        def check_if_path_or_net_string(netlist: PathType):
            cdl_suffix = ".cdl"
            spice_suffix = ".spice"
            if netlist is None:
                raise ValueError("Path to cdl (netlist) must be provided!")
            check_suffix = str(netlist).endswith(cdl_suffix) or str(netlist).endswith(spice_suffix)
            return check_suffix # True for cdl and spice, False if net passed
        
        def extract_design_name_from_netlist(file_path: str):
            """ Extracts the design name from the netlist file (found after the final .ends statement in the netlist file)"""
            with open(file_path, 'r') as file:
                lines = file.readlines()
            last_ends_line = None
            for line in lines:
                if line.strip().startswith(".ends"):
                    last_ends_line = line.strip()

            if last_ends_line:
                parts = last_ends_line.split()
                if len(parts) > 1:
                    return parts[1]
                else:
                    return None
                
        def check_command_exists(command: str):
            """ Check if a command exists in the system """
            result = subprocess.run(f"command -v {command}", shell=True, capture_output=True, text=True)
            return result.returncode == 0
        
        def modify_design_name_in_cdl(netlist, design_name):
            design_name_from_cdl = extract_design_name_from_netlist(netlist)
            if design_name_from_cdl is not None and design_name_from_cdl == design_name:
                print(f"Design name from CDL file: {design_name_from_cdl} matches the design name: {design_name}")
            else:
                # replace all occurences of design_name_from_cdl with design_name in the cdl file
                with open(netlist, 'r') as file:
                    filedata = file.read()
                newdata = filedata.replace(design_name_from_cdl, design_name)
                with open(netlist, 'w') as file:
                    file.write(newdata)
        
        def write_spice(input_cdl, output_spice, lvs_schematic_ref_file):
        # create {design_name}.spice
            with open(input_cdl, 'r') as file:
                lines = file.readlines()
                with open(output_spice, 'w') as file2:
                    sky130_spice_path = Path(lvs_schematic_ref_file).resolve()
                    file2.write(f".include {sky130_spice_path}\n")
                    # write the rest of the lines
                    for line in lines:
                        file2.write(line)
        
        if not check_command_exists("netgen"):
            raise RuntimeError("Netgen not found in the system")
        if not check_command_exists("magic"):
            raise RuntimeError("Magic not found in the system")
        os.environ['DESIGN_NAME'] = design_name
        # results go to a tempdirectory 
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir_path = Path(temp_dir).resolve()
            self.pdk_files['temp_dir'] = temp_dir_path
            print("using user specified pdk_root, will search for required files in the specified directory")
            self.pdk_files['pdk_root'] = pdk_root 
            
            lvsmag_path = temp_dir_path / f"{design_name}_lvsmag.spice"
            pex_path = temp_dir_path / f"{design_name}_pex.spice"
            sim_path = temp_dir_path / f"{design_name}_sim.spice"
            spice_path = temp_dir_path / f"{design_name}.spice"
            netlist_from_comp = temp_dir_path / f"{design_name}.cdl"
            gds_path = temp_dir_path / f"{design_name}.gds"
            report_path = temp_dir_path / f"{design_name}_lvs.rpt"
            
            if isinstance(layout, Component):
                layout.write_gds(str(gds_path))
                if netlist is None:
                    netlist = layout.info['netlist'].generate_netlist()
                    with open(str(netlist_from_comp), 'w') as f:
                        f.write(netlist)
                else: 
                    if check_if_path_or_net_string(netlist):
                        shutil.copy(netlist, str(netlist_from_comp))
                    else: 
                        with open(str(netlist_from_comp), 'w') as f:
                            f.write(netlist)
            elif isinstance(layout, PathType):            
                shutil.copy(layout, str(gds_path))
                if netlist is None:
                    raise ValueError("Path to cdl (netlist) must be provided if only gds file is provided! Provide Component alternatively!")
                else:
                    if check_if_path_or_net_string(netlist):
                        shutil.copy(netlist, str(netlist_from_comp))
                    else: 
                        with open(str(netlist_from_comp), 'w') as f:
                            f.write(netlist)
                    
            modify_design_name_in_cdl(str(netlist_from_comp), design_name)
            
            lvsschemref_file = self.pdk_files['lvs_schematic_ref_file'] if lvs_schematic_ref_file is None else lvs_schematic_ref_file
        
            
            write_spice(str(netlist_from_comp), str(spice_path), lvsschemref_file)
            
            magic_script_content = f"""
gds flatglob *\\$\\$*
gds read {gds_path}
load {design_name}

select top cell
extract all
ext2spice lvs
ext2spice -o {str(lvsmag_path)}
load {design_name}
extract all
ext2spice lvs
ext2spice rthresh 0
ext2spice cthresh 0
ext2spice -o {str(pex_path)}
load {design_name}
extract all
ext2spice cthresh 0
ext2spice -o {str(sim_path)}
exit
"""

            with tempfile.NamedTemporaryFile(mode='w', delete=False) as magic_script_file:
                magic_script_file.write(magic_script_content)
                magic_script_path = magic_script_file.name
            
            try:
                
                magicrc_file = self.pdk_files['magic_drc_file'] if magic_drc_file is None else magic_drc_file
                magic_cmd = f"bash -c 'magic -rcfile {magicrc_file} -noconsole -dnull < {magic_script_path}'",
                magic_subproc = subprocess.run(
                    magic_cmd, 
                    shell=True,
                    check=True,
                    capture_output=True
                )
                
                magic_subproc_code = magic_subproc.returncode
                magic_subproc_out = magic_subproc.stdout.decode('utf-8')
                print(magic_subproc_out)
                
                lvssetup_file = self.pdk_files['lvs_setup_tcl_file'] if lvs_setup_tcl_file is None else lvs_setup_tcl_file 
                netgen_command = f'netgen -batch lvs "{str(lvsmag_path)} {design_name}" "{str(spice_path)} {design_name}" {lvssetup_file} {str(report_path)}'
                
                netgen_subproc = subprocess.run(
                    netgen_command,
                    shell=True,
                    check=True, 
                    capture_output=True
                )
                netgen_subproc_code = netgen_subproc.returncode
                netgen_subproc_out = netgen_subproc.stdout.decode('utf-8')
                print(netgen_subproc_out)
                
                if netgen_subproc_code == 0 and magic_subproc_code == 0:
                    print("LVS run succeeded, writing report...")
                else:
                    raise ValueError("LVS run failed")
            
            finally: 
                os.remove(magic_script_path)
                if os.path.exists(f'{design_name}.ext'):
                    os.remove(f'{design_name}.ext')
                # copy the report from the temp directory to the specified location
                if report_handling[0]:
                    shutil.copy(report_path, report_handling[1])
                    
    
    @validate_arguments
    def has_required_glayers(self, layers_required: list[str]):
        """Raises ValueError if any of the generic layers in layers_required: list[str]
        are not mapped to anything in the pdk.glayers dictionary
        also checks that the values in the glayers dictionary map to real Pdk layers"""
        for layer in layers_required:
            if layer not in self.glayers:
                raise ValueError(
                    f"{layer!r} not in self.glayers {list(self.glayers.keys())}"
                )
            if isinstance(self.glayers[layer], str):
                self.validate_layers([self.glayers[layer]])
            elif not isinstance(self.glayers[layer], tuple):
                raise TypeError("glayer mapped value should be str or tuple[int,int]")


    @validate_arguments
    def layer_to_glayer(self, layer: tuple[int, int]) -> str:
        """if layer provided corresponds to a glayer, will return a glayer
        else will raise an exception
        takes layer as a tuple(int,int)"""
        # lambda for finding last matching key in dict from val
        find_last = lambda val, d: [x for x, y in d.items() if y == val].pop()
        if layer in self.glayers.values():
            return find_last(layer, self.glayers)
        elif self.layers is not None:
            # find glayer verfying presence along the way
            pdk_real_layers = self.layers.values()
            if layer in pdk_real_layers:
                layer_name = find_last(layer, self.layers)
                if layer_name in self.glayers.values():
                    glayer_name = find_last(layer_name, self.glayers)
                else:
                    raise ValueError("layer does not correspond to a glayer")
            else:
                raise ValueError("layer is not a layer present in the pdk")
            return glayer_name
        else:
            raise ValueError("layer might not be a layer present in the pdk")

    # TODO: implement LayerSpec type
    @validate_arguments
    def get_glayer(self, layer: str) -> Layer:
        """Returns the pdk layer from the generic layer name"""
        direct_mapping = self.glayers[layer]
        if isinstance(direct_mapping, tuple):
            return direct_mapping
        else:
            return self.get_layer(direct_mapping)

    @validate_arguments
    def get_grule(
        self, glayer1: str, glayer2: Optional[str] = None, return_decimal = False
    ) -> dict[StrictStr, Union[float,Decimal]]:
        """Returns a dictionary describing the relationship between two layers
        If one layer is specified, returns a dictionary with all intra layer rules"""
        if glayer1 not in MappedPDK.valid_glayers:
            raise ValueError("get_grule, " + str(glayer1) + " not valid glayer")
        # decide if two or one inputs and set rules_dict accordingly
        rules_dict = None
        if glayer2 is not None:
            if glayer2 not in MappedPDK.valid_glayers:
                raise ValueError("get_grule, " + str(glayer2) + " not valid glayer")
            rules_dict = self.grules.get(glayer1, dict()).get(glayer2)
            if rules_dict is None or rules_dict == {}:
                rules_dict = self.grules.get(glayer2, dict()).get(glayer1)
        else:
            glayer2 = glayer1
            rules_dict = self.grules.get(glayer1, dict()).get(glayer1)
        # error check, convert type, and return
        if rules_dict is None or rules_dict == {}:
            raise NotImplementedError(
                "no rules found between " + str(glayer1) + " and " + str(glayer2)
            )
        for rule in rules_dict:
            if type(rule) == float and return_decimal:
                rules_dict[rule] = Decimal(str(rule))
        return rules_dict

    @classmethod
    def is_routable_glayer(cls, glayer: StrictStr):
        return any(hint in glayer for hint in ["met", "active", "poly"])

    # TODO: implement
    @classmethod
    def from_gf_pdk(
        cls,
        gfpdk: Pdk,
        **kwargs
    ):
        """Construct a mapped pdk from an existing pdk and the extra parts of MappedPDK
        grid is the grid size in nm"""
        # input type and value validation
        if not isinstance(gfpdk, Pdk):
            raise TypeError("from_gf_pdk: gfpdk arg only accepts GDSFactory pdk type")
        # create argument dictionary
        passargs = dict()
        # pdk args
        passargs["name"]=gfpdk.name
        #passargs["cross_sections"]=gfpdk.cross_sections
        #passargs["cells"]=gfpdk.cells
        #passargs["symbols"]=gfpdk.symbols
        #passargs["default_symbol_factory"]=gfpdk.default_symbol_factory
        #passargs["containers"]=gfpdk.containers
        #passargs["base_pdk"]=gfpdk.base_pdk
        #passargs["default_decorator"]=gfpdk.default_decorator
        passargs["layers"]=gfpdk.layers
        #passargs["layer_stack"]=gfpdk.layer_stack
        #passargs["layer_views"]=gfpdk.layer_views#??? layer view broken???
#        passargs["layer_transitions"]=gfpdk.layer_transitions
#        passargs["sparameters_path"]=gfpdk.sparameters_path
#        passargs["modes_path"]=gfpdk.modes_path
#        passargs["interconnect_cml_path"]=gfpdk.interconnect_cml_path
#        passargs["warn_off_grid_ports"]=gfpdk.warn_off_grid_ports
#        passargs["constants"]=gfpdk.constants
#        passargs["materials_index"]=gfpdk.materials_index
#        passargs["routing_strategies"]=gfpdk.routing_strategies
#        passargs["circuit_yaml_parser"]=gfpdk.circuit_yaml_parser
#        passargs["gds_write_settings"]=gfpdk.gds_write_settings
#        passargs["oasis_settings"]=gfpdk.oasis_settings
#        passargs["cell_decorator_settings"]=gfpdk.cell_decorator_settings
#        passargs["bend_points_distance"]=gfpdk.bend_points_distance
        # MappedPDK args override existing args
        passargs.update(kwargs)
        # create and return MappedPDK
        mappedpdk = MappedPDK(**passargs)
        return mappedpdk

    # util methods
    @validate_arguments
    def util_max_metal_seperation(self, metal_levels: Union[list[int],list[str], str, int] = range(1,6)) -> float:
        """returns the maximum of the min_seperation rule for all layers specfied
        although the name of this function is util_max_metal_seperation, layers do not have to be metals
        you can specify non metals by using metal_levels=list of glayers
        if metal_levels is list of int, integers are converted to metal levels
        if a single int is provided, all metals below and including that int level are considerd
        by default this function returns the maximum metal seperation of metals1-5
        """
        if type(metal_levels)==int:
            metal_levels = range(1,metal_levels+1)
        metal_levels = metal_levels if isinstance(metal_levels,Iterable) else [metal_levels]
        if len(metal_levels)<1:
            raise ValueError("metal levels cannot be empty list")
        if type(metal_levels[0])==int:
            metal_levels = [f"met{i}" for i in metal_levels]
        sep_rules = list()
        for met in metal_levels:
            sep_rules.append(self.get_grule(met)["min_separation"])
        return self.snap_to_2xgrid(max(sep_rules))

    @validate_arguments
    def snap_to_2xgrid(self, dims: Union[list[Union[float,Decimal]], Union[float,Decimal]], return_type: Literal["decimal","float","same"]="float", snap4: bool=False) -> Union[list[Union[float,Decimal]], Union[float,Decimal]]:
        """snap all numbers in dims to double the grid size.
        This is useful when a generator accepts a size or dimension argument
        because there is a chance the cell may be centered (resulting in off grid components)
        args:
        dims = a list OR single number specifying the dimensions to snap to grid
        return_type = return a decimal, float, or the same type that was passed to the function
        snap4: snap to 4xgrid (Defualt false)
        """
        dims = dims if isinstance(dims, Iterable) else [dims]
        dimtype_in = type(dims[0])
        dims = [Decimal(str(dim)) for dim in dims] # process in decimals
        grid = 2 * Decimal(str(self.grid_size))
        grid = grid if grid else Decimal('0.001')
        grid = 2*grid if snap4 else grid
        # snap dims to grid
        snapped_dims = list()
        for dim in dims:
            snapped_dim = grid * (dim / grid).quantize(1, rounding=ROUND_UP)
            snapped_dims.append(snapped_dim)
        # convert to correct type
        if return_type=="float" or (return_type=="same" and dimtype_in==float):
            snapped_dims = [float(snapped_dim) for snapped_dim in snapped_dims]
        # correctly return list or single element
        return snapped_dims[0] if len(snapped_dims)==1 else snapped_dims




from glayout.flow.routing.straight_route import straight_route
from glayout.flow.routing.L_route import L_route
from glayout.flow.routing.c_route import c_route
from glayout.flow.primitives.fet import nmos, pmos
from glayout.flow.primitives.via_gen import via_stack, via_array
from glayout.flow.pdk.util.comp_utils import evaluate_bbox, movex, move, movey, align_comp_to_port, prec_ref_center, center_to_edge_distance
from glayout.flow.pdk.util.port_utils import set_port_orientation, rename_ports_by_orientation, create_private_ports
from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.primitives.guardring import tapring
from gdsfactory.components import rectangle
from typing import Union, Optional
from itertools import product

from gdsfactory import Component


def common_centroid_ab_ba(
	pdk: MappedPDK,
	width: float = 3,
	fingers: int = 4,
	length: Optional[float] = None,
	n_or_p_fet: bool = True,
	rmult: int = 1,
	dummy: Union[bool, tuple[bool, bool]] = True,
	substrate_tap: bool=True
) -> Component:
    """create a comcentroid with 2 transistors placed in two rows with common centroid place. Sources are shorted
    width = width of the transistors
    fingers = number of fingers in the transistors (must be 2 or more)
    length = length of the transistors, None or 0 means use min length
    short_source = if true connects source of both transistors
    n_or_p_fet = if true the comcentroid is made of nfets else it is made of pfets
    substrate_tap: if true place a tapring around the comcentroid (connects on met1)
    """
    # TODO: error checking
    pdk.activate()
    comcentroid = Component()
    # create transistors
    well = None
    if isinstance(dummy, bool):
        dummy = (dummy, dummy)
    if n_or_p_fet:
        fetL = nmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(dummy[0], False),with_dnwell=False,with_substrate_tap=False,rmult=rmult)
        fetR = nmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(False,dummy[1]),with_dnwell=False,with_substrate_tap=False,rmult=rmult)
        well, sdglayer = "pwell", "n+s/d"
    else:
        fetL = pmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(dummy[0], False),dnwell=False,with_substrate_tap=False,rmult=rmult)
        fetR = pmos(pdk, width=width, fingers=fingers,length=length,multipliers=1,with_tie=False,with_dummy=(False,dummy[1]),dnwell=False,with_substrate_tap=False,rmult=rmult)
        well, sdglayer = "nwell", "p+s/d"
    fetRdims = evaluate_bbox(fetR.flatten().remove_layers(layers=[pdk.get_glayer(well)]))
    fetLdims = evaluate_bbox(fetL.flatten().remove_layers(layers=[pdk.get_glayer(well)]))
    # place and flip top transistors such that the drains of bottom and top point towards eachother
    a_topl = comcentroid << fetL
    a_topl = rename_ports_by_orientation(a_topl.mirror_y())
    b_topr = comcentroid << fetR
    b_topr = rename_ports_by_orientation(b_topr.mirror_y())
    a_botr = comcentroid << fetR
    b_botl = comcentroid << fetL
    prec_ref_center(a_topl, snapmov2grid=True)
    prec_ref_center(b_topr, snapmov2grid=True)
    prec_ref_center(a_botr, snapmov2grid=True)
    prec_ref_center(b_botl, snapmov2grid=True)
    # setup for routing (need viadims to know how far to seperate transistors)
    glayer1 = pdk.layer_to_glayer(a_topl.ports["multiplier_0_drain_E"].layer)
    glayer2 = glayer1[0:-1] + str(int(glayer1[-1])+1)
    glayer0 = glayer1[0:-1] + str(int(glayer1[-1])-1)
    g1g2via = via_stack(pdk,glayer1,glayer2)
    g0g1via = via_stack(pdk,glayer0,glayer1)
    # move transistors into position
    min_spacing_y = pdk.snap_to_2xgrid(1*(g1g2via.ysize - pdk.get_grule(glayer1)["min_width"])+pdk.get_grule(glayer1)["min_separation"])
    extra_g1g2_spacing = pdk.snap_to_2xgrid(max(pdk.get_grule(glayer2)["min_separation"]-pdk.get_grule(glayer1)["min_separation"],0))
    min_spacing_y += extra_g1g2_spacing
    min_spacing_x = 3*pdk.get_grule(glayer1)["min_separation"] + 2*g0g1via.xsize - 2*pdk.get_grule("active_diff",sdglayer)["min_enclosure"]
    min_spacing_x = pdk.snap_to_2xgrid(max(min_spacing_x, pdk.get_grule(sdglayer)["min_separation"]))
    a_topl.movex(0-fetLdims[0]/2-min_spacing_x/2).movey(pdk.snap_to_2xgrid(fetRdims[1]/2+min_spacing_y/2))
    b_topr.movex(fetLdims[0]/2+min_spacing_x/2).movey(pdk.snap_to_2xgrid(fetLdims[1]/2+min_spacing_y/2))
    a_botr.movex(fetLdims[0]/2+min_spacing_x/2).movey(pdk.snap_to_2xgrid(-fetLdims[1]/2-min_spacing_y/2))
    b_botl.movex(0-fetLdims[0]/2-min_spacing_x/2).movey(pdk.snap_to_2xgrid(-fetLdims[1]/2-min_spacing_y/2))
    comcentroid.add_padding(default=0,layers=[pdk.get_glayer(well)])
    # if substrate tap place substrate tap, and route dummy to substrate tap
    if substrate_tap:
        tapref = comcentroid << tapring(pdk,evaluate_bbox(comcentroid,padding=1))#,horizontal_glayer="met1")
        comcentroid.add_ports(tapref.get_ports_list(),prefix="tap_")
        try:
            comcentroid<<straight_route(pdk,a_topl.ports["multiplier_0_dummy_L_gsdcon_top_met_W"],comcentroid.ports["tap_W_top_met_W"],glayer2="met1")
        except KeyError:
            pass
        try:
            comcentroid<<straight_route(pdk,b_topr.ports["multiplier_0_dummy_R_gsdcon_top_met_W"],comcentroid.ports["tap_E_top_met_E"],glayer2="met1")
        except KeyError:
            pass
        try:
            comcentroid<<straight_route(pdk,b_botl.ports["multiplier_0_dummy_L_gsdcon_top_met_W"],comcentroid.ports["tap_W_top_met_W"],glayer2="met1")
        except KeyError:
            pass
        try:
            comcentroid<<straight_route(pdk,a_botr.ports["multiplier_0_dummy_R_gsdcon_top_met_W"],comcentroid.ports["tap_E_top_met_E"],glayer2="met1")
        except KeyError:
            pass
    # correct pwell place, add ports, flatten, and return
    comcentroid.add_ports(a_topl.get_ports_list(),prefix="tl_")
    comcentroid.add_ports(b_topr.get_ports_list(),prefix="tr_")
    comcentroid.add_ports(b_botl.get_ports_list(),prefix="bl_")
    comcentroid.add_ports(a_botr.get_ports_list(),prefix="br_")
    # route asrc to asrc
    vsrca1 = comcentroid << g1g2via
    vsrca2 = comcentroid << g1g2via
    align_comp_to_port(vsrca1,movey(comcentroid.ports["tl_multiplier_0_drain_W"],-extra_g1g2_spacing),alignment=("right","bottom"))
    align_comp_to_port(vsrca2,movey(comcentroid.ports["br_multiplier_0_drain_W"],extra_g1g2_spacing),alignment=("right","top"))
    comcentroid << L_route(pdk, movey(vsrca1.ports["top_met_W"],extra_g1g2_spacing), vsrca2.ports["top_met_N"])
    # route bsrc to bsrc
    vsrcb1 = comcentroid << g1g2via
    vsrcb2 = comcentroid << g1g2via
    align_comp_to_port(vsrcb1,comcentroid.ports["tr_multiplier_0_drain_E"],alignment=("left","bottom"))
    align_comp_to_port(vsrcb2,comcentroid.ports["bl_multiplier_0_drain_E"],alignment=("left","top"))
    intermediate_port = comcentroid.ports["bl_multiplier_0_source_E"].copy()
    intermediate_port.layer = pdk.get_glayer(pdk.layer_to_glayer(vsrcb1.ports["top_met_E"].layer))
    comcentroid << L_route(pdk, vsrcb1.ports["top_met_S"], intermediate_port)
    comcentroid << L_route(pdk,intermediate_port, vsrcb2.ports["top_met_S"])
    # route adrain to adrain
    vdraina1 = comcentroid << g0g1via # first via
    align_comp_to_port(vdraina1, comcentroid.ports[f"tl_multiplier_0_row0_col{fingers-1}_rightsd_top_met_N"],alignment=("right","top"))
    align_comp_to_port(vdraina1, comcentroid.ports["tl_multiplier_0_drain_E"],alignment=("right","none"))
    vdraina1.movex(pdk.get_grule(glayer1)["min_separation"])
    comcentroid << straight_route(pdk, vdraina1.ports["top_met_W"],comcentroid.ports["tr_multiplier_0_leftsd_top_met_E"],glayer2=glayer1)
    vdraina2 = comcentroid << g0g1via # second via
    align_comp_to_port(vdraina2, comcentroid.ports["tl_multiplier_0_drain_E"],alignment=("right","c"))
    vdraina2.movex(pdk.get_grule(glayer1)["min_separation"])
    vdraina2_mdprt = movex(vdraina2.ports["bottom_met_S"],pdk.get_grule("met2","via1")["min_enclosure"])
    vdraina2_mdprt.width = vdraina2_mdprt.width + 2*pdk.get_grule("met2","via1")["min_enclosure"]
    comcentroid << straight_route(pdk, vdraina2_mdprt, vdraina1.ports["bottom_met_N"])
    comcentroid << L_route(pdk, vdraina2.ports["top_met_N"],comcentroid.ports["bl_multiplier_0_source_E"])
    # route bdrain to bdrain
    vdrainb1 = comcentroid << g0g1via # first via
    align_comp_to_port(vdrainb1, comcentroid.ports["br_multiplier_0_leftsd_top_met_N"],alignment=("left","bottom"))
    align_comp_to_port(vdrainb1, comcentroid.ports["br_multiplier_0_drain_W"],alignment=("left","none"))
    vdrainb1.movex(-pdk.get_grule(glayer1)["min_separation"])
    # TODO: fix slight overhang (both this one and the adrain->bdrain)
    comcentroid << straight_route(pdk, vdrainb1.ports["top_met_W"],comcentroid.ports["br_multiplier_0_leftsd_top_met_E"],glayer2=glayer1)
    vdrainb2 = comcentroid << g0g1via # second via
    align_comp_to_port(vdrainb2, comcentroid.ports["br_multiplier_0_drain_W"],alignment=("left","c"))
    vdrainb2.movex(-pdk.get_grule(glayer1)["min_separation"])
    vdrainb2_mdprt = movex(vdrainb2.ports["bottom_met_N"],-pdk.get_grule("met2","via1")["min_enclosure"])
    vdrainb2_mdprt.width = vdrainb2_mdprt.width + 2*pdk.get_grule("met2","via1")["min_enclosure"]
    comcentroid << straight_route(pdk, vdrainb2_mdprt, vdrainb1.ports["bottom_met_S"])
    comcentroid << L_route(pdk, vdrainb2.ports["top_met_N"],comcentroid.ports["tl_multiplier_0_source_E"])
    # agate to agate
    gate2rt_sep = pdk.get_grule(glayer2)["min_separation"]
    vgatea1 = comcentroid << g1g2via# first via
    align_comp_to_port(vgatea1,comcentroid.ports["tl_multiplier_0_gate_E"],alignment=("right","bottom"))
    vgatea2 = comcentroid << g1g2via# second via
    align_comp_to_port(vgatea2,comcentroid.ports["br_multiplier_0_gate_S"],alignment=("right","bottom"))
    vgatea2.movey(-gate2rt_sep)
    comcentroid << straight_route(pdk, vgatea2.ports["bottom_met_S"], comcentroid.ports["br_multiplier_0_gate_N"])
    g1extension = pdk.util_max_metal_seperation()+pdk.snap_to_2xgrid(comcentroid.ports["tr_multiplier_0_plusdoped_E"].center[0] - vgatea2.ports["top_met_E"].center[0])
    cext1 = comcentroid << c_route(pdk, vgatea2.ports["top_met_E"], vgatea1.ports["top_met_E"], cglayer=glayer2, extension=g1extension)
    comcentroid.add_ports(ports=cext1.get_ports_list(),prefix="A_gate_route_")
    # bgate to bgate
    vgateb1 = comcentroid << g1g2via# first via
    align_comp_to_port(vgateb1,comcentroid.ports["bl_multiplier_0_gate_E"],alignment=("right","top"))
    vgateb2 = comcentroid << g1g2via# second via
    align_comp_to_port(vgateb2,comcentroid.ports["tr_multiplier_0_gate_S"],alignment=("right","top"))
    vgateb2.movey(gate2rt_sep)
    comcentroid << straight_route(pdk, vgateb2.ports["bottom_met_N"], comcentroid.ports["tr_multiplier_0_gate_S"])
    g2extension = pdk.util_max_metal_seperation()+pdk.snap_to_2xgrid(abs(comcentroid.ports["tl_multiplier_0_plusdoped_W"].center[0] - vgateb1.ports["top_met_W"].center[0]))
    cext2 = comcentroid << c_route(pdk, vgateb2.ports["top_met_W"], vgateb1.ports["top_met_W"], cglayer=glayer2, extension=g2extension)
    comcentroid.add_ports(ports=cext2.get_ports_list(),prefix="B_gate_route_")
    # create better toplevel ports
    b_drainENS = comcentroid << straight_route(pdk, comcentroid.ports["tr_multiplier_0_drain_E"], movex(cext1.ports["con_N"],cext2.ports["con_N"].width/2+pdk.util_max_metal_seperation()), glayer2=glayer1)
    a_drainENS = comcentroid << straight_route(pdk, comcentroid.ports["br_multiplier_0_drain_E"], movex(cext1.ports["con_N"],cext2.ports["con_N"].width/2+pdk.util_max_metal_seperation()), glayer2=glayer1)
    b_sourceENS = comcentroid << straight_route(pdk, comcentroid.ports["tr_multiplier_0_source_E"], movex(cext1.ports["con_N"],cext2.ports["con_N"].width/2+pdk.util_max_metal_seperation()), glayer2=glayer1)
    a_sourceENS = comcentroid << straight_route(pdk, comcentroid.ports["br_multiplier_0_source_E"], movex(cext1.ports["con_N"],cext2.ports["con_N"].width/2+pdk.util_max_metal_seperation()), glayer2=glayer1)
    b_drainW = comcentroid << straight_route(pdk, comcentroid.ports["bl_multiplier_0_drain_W"], movex(cext2.ports["con_N"],-cext2.ports["con_N"].width/2-pdk.util_max_metal_seperation()), glayer2=glayer1)
    a_drainW = comcentroid << straight_route(pdk, comcentroid.ports["tl_multiplier_0_drain_W"], movex(cext2.ports["con_N"],-cext2.ports["con_N"].width/2-pdk.util_max_metal_seperation()), glayer2=glayer1)
    b_sourceW = comcentroid << straight_route(pdk, comcentroid.ports["bl_multiplier_0_source_W"], movex(cext2.ports["con_N"],-cext2.ports["con_N"].width/2-pdk.util_max_metal_seperation()), glayer2=glayer1)
    a_sourceW = comcentroid << straight_route(pdk, comcentroid.ports["tl_multiplier_0_source_W"], movex(cext2.ports["con_N"],-cext2.ports["con_N"].width/2-pdk.util_max_metal_seperation()), glayer2=glayer1)
    # add the ports
    def makeNorS(portin, direction: str):
        mdprt = set_port_orientation(movex(portin.copy(),(-1 if portin.name.endswith("E") else 1)*pdk.snap_to_2xgrid(portin.width/2)),direction)
        mdprt.name = (mdprt.name.strip("EW") + direction.strip().capitalize()).removeprefix("route_")
        return movey(mdprt,(1 if direction.endswith("N") else -1)*pdk.snap_to_2xgrid(mdprt.width/2))
    def addENS(topcomp: Component, straightrouteref, device: str, pin: str):
        # device is A or B and pin is source drain or gate
        eastport = straightrouteref.ports["route_E"].copy()
        eastport.name = eastport.name.removeprefix("route_")
        topcomp.add_ports(ports=[eastport,makeNorS(eastport,"N"),makeNorS(eastport,"S")],prefix=device+"_"+pin+"_")
    addENS(comcentroid,b_drainENS,"B","drain")
    addENS(comcentroid,a_drainENS,"A","drain")
    addENS(comcentroid,b_sourceENS,"B","source")
    addENS(comcentroid,a_sourceENS,"A","source")
    def localportrename(portin):
        portin = portin.copy()
        portin.name = portin.name.removeprefix("route_")
        return portin
    comcentroid.add_ports(ports=[localportrename(b_drainW.ports["route_W"])],prefix="B_drain_")
    comcentroid.add_ports(ports=[localportrename(a_drainW.ports["route_W"])],prefix="A_drain_")
    comcentroid.add_ports(ports=[localportrename(b_sourceW.ports["route_W"])],prefix="B_source_")
    comcentroid.add_ports(ports=[localportrename(a_sourceW.ports["route_W"])],prefix="A_source_")
    # better gate routes
    a_gateE = comcentroid << straight_route(pdk, comcentroid.ports["br_multiplier_0_gate_E"], comcentroid.ports["A_drain_E"])
    b_gateE = comcentroid << straight_route(pdk, comcentroid.ports["tr_multiplier_0_gate_E"], comcentroid.ports["A_drain_E"])
    a_gateW = comcentroid << straight_route(pdk, comcentroid.ports["tl_multiplier_0_gate_W"], comcentroid.ports["A_drain_W"])
    b_gateW = comcentroid << straight_route(pdk, comcentroid.ports["bl_multiplier_0_gate_W"], comcentroid.ports["A_drain_W"])
    comcentroid.add_ports(ports=[localportrename(a_gateE.ports["route_E"])],prefix="A_gate_")
    comcentroid.add_ports(ports=[localportrename(b_gateE.ports["route_E"])],prefix="B_gate_")
    comcentroid.add_ports(ports=[localportrename(a_gateW.ports["route_W"])],prefix="A_gate_")
    comcentroid.add_ports(ports=[localportrename(b_gateW.ports["route_W"])],prefix="B_gate_")
    rename_north_portb = vgateb2.ports["top_met_N"].copy()# add B_gate_N
    rename_north_portb.name = "B_gate_N"
    comcentroid.add_ports(ports=[rename_north_portb])
    rename_south_porta = vgatea2.ports["top_met_S"].copy()# add A_gate_S
    rename_south_porta.name = "A_gate_S"
    comcentroid.add_ports(ports=[rename_south_porta])
    rename_south_portb = vgateb1.ports["top_met_S"].copy()# add B_gate_S
    rename_south_portb.name = "B_gate_S"
    comcentroid.add_ports(ports=[rename_south_portb])
    # rename ports and add private ports for smart route
    comcentroid = rename_ports_by_orientation(comcentroid)
    comcentroid.add_ports(create_private_ports(comcentroid,["".join(prtp) for prtp in product(["A_","B_"],["drain","source","gate"])]))
    comcentroid.info["route_genid"]="common_centroid_ab_ba"
    return comcentroid




import copy
import io
from pathlib import Path
from typing import Optional, Union

import nltk
import glayout.syntaxer.nltk_init_deps
import glayout.syntaxer.dynamic_load
from glayout.syntaxer.relational import GlayoutCode, parse_direction
from glayout.flow.pdk.sky130_mapped import sky130_mapped_pdk

class Session:
    """The session stores all relevant information for producing code from a conversation"""

    generic_prompt = """Place a cell, move a cell, route, create a parameter, or define a variable.
You can also dump code or save this conversation, or enter "help" to see supported syntax in detail
What would you like to do?"""

    def __init__(
        self,
        outputstream: io.IOBase,
        inputstream: Optional[io.IOBase] = None,
        toplvlname: Optional[str] = None,
    ):
        """initialize a conversation and greet the user
        Args:
                outputstream (io.IOBase): used to print outputs
                inputstream (io.IOBase): saved for (optionally) reading in user input, also just provide a string
                NOTE: if input stream not provided, str input must be provided
                toplvlname (str): in string only mode, you can input toplvl name using this arg
        """
        self.inputstream = inputstream
        self.outputstream = outputstream
        # greet the user and get a top level name for the component
        if toplvlname is None:
            if inputstream is None:
                raise RuntimeError(
                    "you must specify AT LEAST one of inputstream or name"
                )
            self.print_to_stream("Hello!")
            self.print_to_stream(
                "Please provide a name for the Component you want to create"
            )
            self.print_to_stream(
                "remember, this will be the name of your top level component: "
            )
            self.name = self.read_from_stream().strip()
        else:
            self.name = str(toplvlname).strip()
        # save a list of responses to recreate this component from a .conv file
        self.conversation_responses = list()
        self.conversation_responses.append(self.name)
        # init the code object
        self.code = GlayoutCode(self.name)
        # create a backup that goes back exactly one call to process_next_input
        self.backup = self.__backup()

    def __backup(self):
        """Produce an exact copy of this class to revert after an exception"""
        newobj = self.__class__
        backup = newobj.__new__(newobj)
        backup.inputstream = self.inputstream
        backup.outputstream = self.outputstream
        backup.conversation_responses = copy.deepcopy(self.conversation_responses)
        backup.code = copy.deepcopy(self.code)
        backup.backup = None
        return backup

    def print_help(self):
        """print help message to self.outputstream"""
        syntaxes = list()
        # place_syntax
        syntaxes.append("place an genidhere called/named compnamehere with paramshere")
        # route_syntax
        syntaxes.append("route between/from port1 and/to port2 using routetype with paramshere")
        # absolute move syntax
        syntaxes.append("move compname [by] (x,y)")
        # relative move
        syntaxes.append("move compname [filler words] direction [filler words] reference_comp [by separation]")
        # import
        import_syntax = "import comp1, comp2 from mod1, and comp3 from some/path/mod.py"
        import_syntax += "\n\tNOTE: imports from glayout only need component name"
        import_syntax += "\n\tNOTE: if mod name is not specified, it is assumed to be the component name"
        syntaxes.append(import_syntax)
        # parameter
        syntaxes.append("create/define [a/an] param_type parameter called/named paramname")
        # variable
        syntaxes.append("create/define [a/an] var_type variable called/named varname =/equal valorexpr")
        syntaxes.append("save/dump [conversation/code] to pathtosaveto\n\tNOTE: save code is assumed if neither conversation nor code are specified")
        syntaxes.append("show\n\tNOTE: This will show the current component in klayout using .show and klive plugin")
        # print all
        self.print_to_stream("\nBelow are valid sentences:")
        for syntax in syntaxes:
            self.print_to_stream(syntax)
        self.print_to_stream()

    def save_to_disk(
        self, savecode: bool = True, savepath: Optional[Union[str, Path]] = None
    ):
        """Save NLP results to disk, either save code or conversation responses.
        Args:
                savemode: bool, default to True==save code, false==save conversation responses
                savepath: Optional (if None, defaults to ./),
                NOTE: None is the same as specifying the current directory "./"
        """
        # figure out which save mode we are in
        suffix = ".py" if savecode else ".convo"
        if savepath is None:
            savepath = Path("./").resolve()
        savepath = Path(savepath.strip()) if isinstance(savepath, str) else savepath
        if savepath.is_dir():
            savepath /= f"{self.code.toplvl_name}{suffix}"
        if savepath.parent.is_dir():
            with savepath.open("w") as savefile:
                if savecode:
                    savefile.write(self.code.get_code())
                else:
                    savefile.writelines([r + "\n" for r in self.conversation_responses])
        else:
            raise FileNotFoundError("the directory you specfied does not exist")

    def print_to_stream(self, toprint: Optional[str] = None):
        """prints to the configured outputstream

        Args:
            toprint (Optional[str]): string to print. If None then prints a newline.
        """
        if toprint is None:
            toprint = ""
        self.outputstream.write(toprint + "\n")

    def read_from_stream(self) -> str:
        """reads user input from the configured inputstream

        Returns:
            str: the next line of user input
        """
        response = self.inputstream.readline()
        # self.conversation_responses.append(response)
        return response

    def __save_response(self, response: str):
        self.conversation_responses.append(response.strip())

    def process_import_sentence(self, text_input: str) -> bool:
        """Will update the internal code table using a sentence which follows the import syntax

        Args:
            text_input (str): user input text
        
        Returns:
            bool: saveresponse
        """
        words = nltk.word_tokenize(text_input)
        imports = text_input.replace("and", ",").replace("import", "").replace("from", "").strip().split(",")
        for modimport in imports:
            if modimport == "" or modimport.isspace():
                continue
            words = modimport.strip().split()
            compname = words[0]
            modpath = words[1] if len(words) > 1 else None
            aliases = [compname]  # TODO implement comp aliasing
            self.code.update_import_table(aliases, compname, modpath)
        return True

    def process_param_or_var_sentence(self, text_input: str) -> bool:
        """Will update the internal code table using a sentence which follows the variable or parameter define syntax

        Args:
            text_input (str): user input text
        
        Returns:
            bool: saveresponse
        """
        words = nltk.word_tokenize(text_input)
        vartype = None
        for word in words:
            word = word.lower().strip()
            if "int" in word:
                vartype = int
            elif "float" in word:
                vartype = float
            elif "bool" in word:
                vartype = bool
            elif "str" in word:
                vartype = str
            elif "tupl" in word:
                vartype = tuple
            if vartype is not None:
                break
        varname = None
        for i, word in enumerate(words):
            if word in ["called", "named"]:
                varname = words[i + 1]
                break
        expr = None
        eqpar = text_input.replace("equal", "=").strip().removesuffix(".")
        expr = eqpar.split("=")[1] if "=" in eqpar else None
        if "parameter" in words:
            self.code.update_parameter_table(varname, vartype, None, None)
        else:  # variable
            self.code.update_variable_table(varname, expr)
        return True

    def process_place_sentence(self, text_input: str) -> bool:
        """Will update the internal code table using a sentence which follows the place syntax

        Args:
            text_input (str): user input text
        
        Returns:
            bool: saveresponse
        """
        words = text_input.strip().split()#nltk.word_tokenize(text_input)
        if "configuration" in text_input:
            raise NotImplementedError("configs not yet implemented")
        # util func
        def split_input_on_compname(text_input:str) -> str:
                """Split a string on the words 'called' or 'named' and retain everything before these keywords.
                Args:
                    text_input (str): The input string.
                Returns:
                    str: The part of the text_input before 'called' or 'named'.
                """
                # look for either named or called in a lower case version of parts
                lc_parts = text_input.lower().split()
                indexCalled, indexNamed = None, None
                if "called" in lc_parts:
                    indexCalled = lc_parts.index("called")
                if "named" in lc_parts:
                    indexNamed = lc_parts.index("named")
                # pick which index to use or raise error if neither
                index = None
                if (indexCalled is not None) and (indexNamed is not None):
                    index = max(indexCalled,indexNamed)
                else:
                    index = indexCalled if (indexCalled is not None) else indexNamed
                if index is None:
                    raise SyntaxError("invalid place syntax, place sentence must include 'called' or 'named' keyword")
                # return everything in the sentence before the "called" or "named" keyword
                parts = text_input.split()
                return " ".join(parts[0:index])
        genid = self.code.find_first_generator_id(split_input_on_compname(text_input))
        comp_name = None
        for i, word in enumerate(words):
            if word in ["called", "named"]:
                comp_name = words[i + 1]
                break
        params = None
        for i, word in enumerate(words):
            if word == "with":
                params = " ".join(words[i:])
                break
        self.code.update_place_table(genid, params, comp_name)
        return True

    def process_route_sentence(self, text_input: str) -> bool:
        """Will update the internal code table using a sentence which follows the route syntax

        Args:
            text_input (str): user input text
        
        Returns:
            bool: saveresponse
        """
        words = nltk.word_tokenize(text_input)
        port1 = None
        port2 = None
        routetype = None
        params = None
        for i, word in enumerate(words):
            if word in ["between", "from"]:
                port1 = words[i + 1] if port1 is None else port1
            elif word in ["from", "and"]:
                port2 = words[i + 1] if port2 is None else port2
            elif word in ["using", "a", "an"]:
                routetype = words[i + 1]
            elif word == "with":
                params = " ".join(words[i:])
        self.code.update_route_table(port1, port2, params, routetype)
        return True

    def process_move_sentence(self, text_input: str) -> bool:
        """Will update the internal code table using a sentence which follows the route syntax

        Args:
            text_input (str): user input text
        
        Returns:
            bool: saveresponse
        """
        words = nltk.word_tokenize(text_input)
        direction = None
        reference_comp = None
        separation = None
        # words[0] is "move" and words[1] is the component name
        for i, word in enumerate(words[2:]):
            # try to find direction until we see it then move on
            if parse_direction(word) is not None and direction is None:
                direction = word
            # once you see a direction, look for a reference_comp
            elif (word=="by" or word=="with") and reference_comp is None:
                reference_comp = words[2:][i-1]
                separation = words[2:][i+1]
                if len(separation)>3 and separation.strip().lower()[0:3]=="sep":
                    separation = words[2:][i+2]
        # you may not find it in the loop, in which case reference_comp is the last word
        if reference_comp is None:
            reference_comp = words[-1]
        # relative move
        if direction is not None:
            spacesepwords = text_input.split()
            self.code.update_move_table("relative", words[1], reference_comp, direction, separation)
        else:  # assume absolute move
            words = nltk.word_tokenize(
                text_input.replace("(", "").replace(")", "").replace("by", "").replace("with","")
            )
            move_distance = words[2]
            if len(words) > 4 and "," in words[3]:
                move_distance = words[2] + "," + words[4]
            self.code.update_move_table("absolute", words[1], move_distance)
        return True

    def show_current_component(self, text_input: str) -> False:
        """displays the current state of the layout with klayout using .show in sky130nm tech
        if the keyword "port" is found within text_input, port tree (with depth 6) will be saved to a file instead
        Args:
            text_input (str): user input text
        Returns:
            False: saveresponse=False
        """
        if "port" in text_input.lower():
            glayout.syntaxer.dynamic_load.printPortTree_glayout_code_cell(sky130_mapped_pdk,self.code.get_code())
        elif "param" in text_input.lower():
            print(*self.code.parameter_table,sep="\n")
        else:
            glayout.syntaxer.dynamic_load.show_glayout_code_cell(sky130_mapped_pdk, self.code.get_code())
        return False
    
    def process_next_input(self, text_input: str) -> bool:
        """main driver for doing things
        returns True if session is ongoing
        """
        self.backup = self.__backup()
        sentences = nltk.sent_tokenize(text_input)
        for sentence in sentences:
            saveresponse = True
            sentence = sentence.strip().removesuffix(".")
            words = nltk.word_tokenize(sentence)
            mode_indicator = words[0].strip().replace("-", "").lower()
            # parse user input
            if mode_indicator[0] == "h":  # help
                saveresponse = False
                self.print_help()
            elif mode_indicator[0] == "i":  # import
                saveresponse = self.process_import_sentence(sentence)
            elif "create" in mode_indicator or "define" in mode_indicator:  # create/define a variable or param
                saveresponse = self.process_param_or_var_sentence(sentence)
            elif mode_indicator[0] == "p":  # place
                saveresponse = self.process_place_sentence(sentence)
            elif mode_indicator[0] == "r":  # route
                saveresponse = self.process_route_sentence(sentence)
            elif mode_indicator[0] == "m":  # move
                saveresponse = self.process_move_sentence(sentence)
            elif mode_indicator[0] == "e":  # dump code and exit
                saveresponse = False
                self.print_to_stream("\n" + self.code.get_code())
                return False
            elif (
                mode_indicator == "save" or mode_indicator == "dump"
            ):  # save the conversation to a .convo file, or code to .py file
                saveresponse = False
                savepath = None
                for i, word in enumerate(words):
                    if word in ["to", "at", "path"]:
                        savepath = "".join(words[i + 1 :])
                        break
                savecode = any(
                    wr in words for wr in ["code", "python", "py", "program"]
                )
                self.save_to_disk(savecode, savepath)
            elif mode_indicator[0] == "s":  # show a component
                saveresponse = self.show_current_component(sentence)
            elif mode_indicator[0] == "#" or mode_indicator[0]=="/": # comment
                saveresponse=True
            else:
                self.print_to_stream("invalid input")
                self.print_to_stream("sentences must begin with either place, route, generate, show, dump, or move")
                saveresponse = False
            # save when needed and return True to continue prompting for input
            if saveresponse:
                self.__save_response(sentence)
            return True



import json
import warnings
from pathlib import Path
from typing import List, Tuple, Union
from langchain_community.document_loaders import DirectoryLoader, TextLoader
from langchain_community.vectorstores import Chroma
#from langchain_community.embeddings import HuggingFaceEmbeddings
from langchain_huggingface import HuggingFaceEmbeddings
from datasets import Dataset
from typing import Optional


def remove_comments_and_empty_lines(input_string: str) -> str:
    """
    Removes all lines starting with a '#' or any empty lines from the input string.
    Args:
        input_string (str): The input string containing multiple lines.
    Returns:
        str: The modified string with comments and empty lines removed.
    """
    # Split the input string into individual lines
    lines = input_string.split('\n')
    # Filter out lines that start with # or are empty
    filtered_lines = str()
    for line in lines:
        remove = line.strip().startswith('#') or line.strip() == ''
        if not remove:
            filtered_lines += line + "\n"
    return filtered_lines


def load_labeled_syntax_data_json(
    json_file_path: Union[str, Path],
    convo_dir_path: Union[str, Path],
    fail_on_error: bool = True,
) -> List[Tuple[str, str]]:
    """Load all labeled data examples from a JSON file by reading LLM prompt
    and extracting strict syntax from corresponding '.convo' files.
    Args:
        json_file_path (str or Path): The path to the JSON file.
        convo_dir_path (str or Path): The path to the directory containing convo files for training
        fail_on_error (bool): if False, issue warnings if an individual example raises an Exception
    Raises:
        FileNotFoundError: If a '.convo' file, the convo_dir_path or json_file_path do not exist
        KeyError: if the JSON file does not correspond to the expected form
            jsonfile[data] must be a list of examples containing an LLMprompt and NLPfilename
    Returns:
        list of tuple: A list of tuples containing information about each example.
            Each tuple consists of:
                - str: The LLM prompt.
                - str: The text found in the NLP file.
    """
    # process the convo dir path
    convo_dir_path = Path(convo_dir_path).resolve()
    if not convo_dir_path.is_dir():
        raise FileNotFoundError(f"convo directory not found {convo_dir_path}")
    # this requires manage_data.py to be in the llm folder
    # Load JSON data
    json_file_path = Path(json_file_path).resolve()
    if not json_file_path.is_file():
        raise FileNotFoundError(f"could not find JSON file {json_file_path}")
    with open(json_file_path, "r") as file:
        data = json.load(file)
    examples = data.get("data")
    if examples is None:
        raise KeyError(
            f"data not found in JSON file {json_file_path}, ensure 'data' keyword is used"
        )
    # loop through examples and append to results
    results = list()
    for example in examples:
        try:
            llm_prompt = example.get("LLMprompt")
            if llm_prompt is None:
                raise KeyError(f"LLMprompt not found in JSON data {example}")
            # get the name of the strict syntax convo corresponding to this example
            nlp_filename = example.get("NLPfilename")
            if nlp_filename is None:
                raise KeyError(f"NLPfilename not found in JSON data {example}")
            convo_filename = f"{nlp_filename.strip()}"
            if not convo_filename.endswith(".convo"):
                convo_filename += ".convo"
            # get the .convo file path and check that it exists
            convo_file_path = Path(convo_dir_path).resolve() / convo_filename
            if not convo_file_path.is_file():
                raise FileNotFoundError(f"Convo file '{convo_file_path}' not found.")
            # Read text from convo file
            with convo_file_path.open("r") as nlp_file:
                nlp_text = remove_comments_and_empty_lines(nlp_file.read())
            results.append((llm_prompt, nlp_text))
        except (FileNotFoundError, KeyError) as faulty_example_error:
            if fail_on_error:
                raise faulty_example_error
            print(f"\n\nan exception was encounterd when processing {example}\n")
            warnings.warn(f"{faulty_example_error}")
    return results


def load_all_labeled_syntax_data_json(
    evaluation: bool = False,
) -> List[Tuple[str, str]]:
    """look in syntax_data folder for any json files
    and run load_labeled_syntax_data_json on those files
    Args:
        evaluation (bool, Optional): if True, only load json files containing keyword "eval", default False
    Returns:
        list[tuple]: all results from running load_labeled_syntax_data_json on all json files
    """
    # Path to the directory containing JSON files
    llm_dir = Path(__file__).resolve().parent
    json_dir = llm_dir / "syntax_data"
    convo_dir = json_dir / "convos"
    if not json_dir.is_dir():
        raise FileNotFoundError(
            "Could not find syntax_data directory within llm subpackage of glayout"
        )
    # Iterate over all JSON files in the directory
    all_results = list()
    for json_file_path in json_dir.glob("*.json"):
        eval_in_name = "eval" in json_file_path.name
        # load evaluation data and evaluation is in name of this json
        if evaluation and eval_in_name:
            all_results.extend(
                load_labeled_syntax_data_json(json_file_path, convo_dir, False)
            )
        # dont load evaluation data and evaluation is not in name of this json
        elif not (evaluation or eval_in_name):
            all_results.extend(
                load_labeled_syntax_data_json(json_file_path, convo_dir, False)
            )
    return all_results


class RAGdb:
    """
    A class to create and manage a vector database for the RAG data using ChromaDB.

    Attributes:
        chroma_client (Client): The ChromaDB client used for managing the vector database.
        collection_name (str): The name of the collection used in ChromaDB.
        collection (Collection): The vector database
    """

    def __init__(self, rag_data_dir: Union[str, Path]):
        """Initializes the RAGdb instance with a ChromaDB collection"""
        # error checking
        rag_data_dir = Path(rag_data_dir).resolve()
        if not rag_data_dir.is_dir():
            raise FileNotFoundError(f"could not find RAG data directory {rag_data_dir}")
        # load RAG data
        self.documents = DirectoryLoader(str(rag_data_dir), glob="*.md").load()
        # create vector db
        embeddings_model_name = "sentence-transformers/all-MiniLM-L6-v2"
        embeddings = HuggingFaceEmbeddings(model_name=embeddings_model_name)
        self.vectordb = Chroma.from_documents(
            documents=self.documents, embedding=embeddings
        )

    def query(self, query_text: str, k: int = 1) -> list:
        """
        Queries the vector database to find the top-k most similar vectors to the given query text.
        Args:
            query_text (str): The text to query.
        Returns:
            List: The list of top-k most similar docs.
        """
        kact = k if k>1 else 2
        rawdocs = self.vectordb.similarity_search(query=query_text, k=kact)
        rawtxt = list()
        for i, doc in enumerate(rawdocs):
            rawtxt.append(doc.page_content)
            if i == kact-1:
                break
        return rawtxt


RAGvecdb = RAGdb(Path(__file__).resolve().parent / "rag_data")


def get_glayout_context() -> str:
    """retrieve the context of syntax_data/GlayoutStrictSyntax.md
    to provide context of Glayout strictsyntax inserted in the prompt

    Returns:
        str: string content of GlayoutStrictSyntax.md
    """
    contextmdfile = (
        Path(__file__).resolve().parent / "syntax_data/GlayoutStrictSyntax.md"
    )
    loader = TextLoader(contextmdfile)
    return loader.load()[0].page_content


def get_prompt_from_template(
    tokenizer,
    ragcontext: str,
    prompt: str,
    strictsyntax: Optional[str] = None,
    return_message: bool=False
) -> str:
    """Generate a structured prompt for translating input text to Glayout strictsyntax.
    Will extract glayout NLPcontext from the get_glayout_context function.

    Args:
        tokenizer: a tokenizer compatible with huggingface, transformers tokenizer class
        ragcontext (str): Contextual information about analog circuit design to aid in translation.
        prompt (str): The input prompt that needs to be converted to Glayout strictsyntax.
        strictsyntax (str, Optional): The strictsyntax command language template to be used.
            if None (default), then only format the prompt (no labeled output strictsyntax)
        return_message (bool, Optional): if True, will not combine the prompt, rather it will return messages format

    Returns:
        str: The generated prompt formatted with the provided context and input data.
    """
    inst_prompt = str()
    glayout_nlp_context = get_glayout_context()
    inst_prompt += f"Below is some context on Glayout strictsyntax:\n{glayout_nlp_context}\n\n"
    #inst_prompt += "Below is context on the circuit"
    #inst_prompt += "convert an example prompt to Glayout strictsyntax\n"
    #inst_prompt += f"{ragcontext}\n\n----\nTRANSLATION TASK\n"
    #inst_prompt += f"Do NOT include the context in your response. Convert the following prompt to Glayout strictsyntax:\n{prompt}"
    inst_prompt += f"Glayout strictsyntax is a electronic circuit layout command language. Convert the following prompt to Glayout strictsyntax:\n{prompt}"
    # unify prompt and return
    messages = [{"role": "user", "content": inst_prompt}]
    # conditionally add label (expected strict syntax output)
    if strictsyntax is not None:
        messages.append({"role": "assistant", "content": strictsyntax})
    if return_message:
        return messages
    return tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)

# pass all prompts through rag before handing training data to llm
def unify_prompt_and_add_context_to_data(tokenizer, data: list, no_label: bool=False) -> list:
    """Enhance prmopts with vectordb contextual information
    constructs a new prompt incorporating this contextual information according to a specified template
    Args:
        tokenizer: a tokenizer compatible with huggingface, transformers tokenizer class
        data (list): A list of tuples, where each tuple is (prompt (str), result (str)) (or a list of prompts (str))
        no_label (bool): set this to true if you pass data in with JUST the prompt (no strictsyntax label)
    Returns:
        list[str]: A list of strings (prompt and strictsyntax have been combined)
    """
    glayout_context = get_glayout_context()
    if no_label:
        data = [(prompt, None) for prompt in data]
    contextualized_prompts = list()
    for prompt, result in data:
        docs = RAGvecdb.query(prompt, 1)
        ragdata = str()
        for doc in docs:
            # ragdata += f"[CONTEXT DOCUMENT NUMBER {i}]\n"
            ragdata += "\n" + doc + "\n"
            # ragdata += f"[/CONTEXT DOCUMENT NUMBER {i}]\n"
        newprompt = get_prompt_from_template(
            tokenizer=tokenizer,
            ragcontext=ragdata,
            prompt=prompt,
            strictsyntax=result
        )
        contextualized_prompts.append(newprompt)
    return contextualized_prompts


def load_preprocessed_data_in_messages_format():
    # get train and evaluation data in a single unified prompt format
    train_examples = load_all_labeled_syntax_data_json()
    eval_examples = load_all_labeled_syntax_data_json(True)
    # train
    train_messages = list()
    for prompt, result in train_examples:
        train_messages.append(get_prompt_from_template(None,None,prompt,result,True))
    train_data = Dataset.from_dict({"messages":train_messages})
    # eval
    eval_messages = list()
    for prompt, result in eval_examples:
        eval_messages.append(get_prompt_from_template(None,None,prompt,result,True))
    eval_data = Dataset.from_dict({"messages":eval_messages})
    return {"train": train_data, "evaluation": eval_data}




def load_preprocessed_pretokenized_data(tokenizer):
    """Wrapper function for full retrival and preprocessing of dataset
    1- Loads raw data from files
    2- adds RAG context to the LLM prompts
    3- unifies the prompt and strictsyntax output
    4- converts the dataset to Dataset type with example, input_ids, and attention_mask
    Args:
        tokenizer: LM tokenizer which follows hugging face tokenizer model
    Returns:
        dataset in the exact format needed for training
    """
    # get train and evaluation data in a single unified prompt format
    train_examples = unify_prompt_and_add_context_to_data(tokenizer, load_all_labeled_syntax_data_json())
    eval_examples = unify_prompt_and_add_context_to_data(tokenizer, load_all_labeled_syntax_data_json(True))
    # tokenize the prompts
    train_data = tokenizer(train_examples, padding=True)
    eval_data = tokenizer(eval_examples, padding=True)
    # use from dict method to convert to hugging face dataset
    train_data = Dataset.from_dict(
        {
            "example": train_examples,
            "input_ids": train_data.input_ids,
            "attention_mask": train_data.attention_mask,
        }
    )
    eval_data = Dataset.from_dict(
        {
            "example": eval_examples,
            "input_ids": eval_data.input_ids,
            "attention_mask": eval_data.attention_mask,
        }
    )
    # combine to create the final dataset
    return {"train": train_data, "evaluation": eval_data}



"""
usage: from gf180_mapped import gf180_mapped_pdk
"""

from ..gf180_mapped.grules import grulesobj
from ..mappedpdk import MappedPDK, SetupPDKFiles
from pathlib import Path


#LAYER["fusetop"]=(75, 0)
LAYER = {
    "metal5": (81, 0),
    "via4": (41, 0),
    "metal4": (46, 0),
    "via3": (40, 0),
    "metal3": (42, 0),
    "via2": (38, 0),
    "metal2": (36, 0),
    "via1": (35, 0),
    "metal1": (34, 0),
    "contact": (33, 0),
    "poly2": (30, 0),
    "comp": (22, 0),
    "nplus": (32, 0),
    "pplus": (31, 0),
    "nwell": (21, 0),
    "lvpwell": (204, 0),
    "dnwell": (12, 0),
    "CAP_MK": (117, 5)
}

gf180_glayer_mapping = {
    "met5": "metal5",
    "via4": "via4",
    "met4": "metal4",
    "via3": "via3",
    "met3": "metal3",
    "via2": "via2",
    "met2": "metal2",
    "via1": "via1",
    "met1": "metal1",
    "mcon": "contact",
    "poly": "poly2",
    "active_diff": "comp",
    "active_tap": "comp",
    "n+s/d": "nplus",
    "p+s/d": "pplus",
    "nwell": "nwell",
    "pwell": "lvpwell",
    "dnwell": "dnwell",
    "capmet": "CAP_MK"
}

# note for DRC, there is mim_option 'A'. This is the one configured for use

gf180_lydrc_file_path = Path(__file__).resolve().parent / "gf180mcu_drc.lydrc"

openfasoc_dir = Path(__file__).resolve().parent.parent.parent.parent.parent.parent.parent
pdk_root = Path('/usr/bin/miniconda3/share/pdk/')
lvs_schematic_ref_file = openfasoc_dir / "common" / "platforms" / "gf180osu9t" / "cdl" / "gf180mcu_osu_sc_9T.spice"
magic_drc_file = pdk_root / "gf180mcuC" / "libs.tech" / "magic" / "gf180mcuC.magicrc"
lvs_setup_tcl_file = pdk_root / "gf180mcuC" / "libs.tech" / "netgen" / "gf180mcuC_setup.tcl"
temp_dir = None


pdk_files = SetupPDKFiles(
    pdk_root=pdk_root,
    klayout_drc_file=gf180_lydrc_file_path,
    lvs_schematic_ref_file=lvs_schematic_ref_file,
    lvs_setup_tcl_file=lvs_setup_tcl_file,
    magic_drc_file=magic_drc_file,
    temp_dir=temp_dir,
    pdk='gf180'
).return_dict_of_files()

gf180_mapped_pdk = MappedPDK(
    name="gf180",
    glayers=gf180_glayer_mapping,
	models={
        'nfet': 'nfet_03v3',
		'pfet': 'pfet_03v3',
		'mimcap': 'mimcap_1p0fF'
    },
    layers=LAYER,
    pdk_files=pdk_files,
    grules=grulesobj,
)

# configure the grid size and other settings
gf180_mapped_pdk.gds_write_settings.precision = 5*10**-9
gf180_mapped_pdk.cell_decorator_settings.cache=False



from glayout.flow.primitives.fet import pmos, nmos
from glayout.flow.routing.c_route import c_route
from glayout.flow.routing.L_route import L_route
from glayout.flow.routing.straight_route import straight_route
from glayout.flow.routing.smart_route import smart_route
from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.primitives.guardring import tapring
from glayout.flow.pdk.util.comp_utils import evaluate_bbox, add_ports_perimeter
from gdsfactory.component import Component
from gdsfactory.cell import cell 
from typing import Optional

@cell
def resistor(
    pdk: MappedPDK,
    width: float = 5,
    length: float = 1,
    num_series: int = 1,
    with_substrate_tap: bool = False,
    with_tie: bool = False,
    with_dnwell: bool = False,
    rmult: Optional[int] = None,
    multipliers: int = 1,
    substrate_tap_layers: tuple[str, str] = ("met2", "met1"),
    tie_layers: tuple[str, str] = ("met2", "met1"),
) -> Component:
    """This cell represents a diode connected pfet which acts as a programmable resistor. The small signal resistance is modelled by (1/gm)||r_0, where gm is the fet's transconductance and r0 is the small signal output impedance. The cell can be instantiated with any choice of width and length. The number of resistors connected in series can be controlled using numseries (**note: they will be placed in a single line, so area saving has to be manually handled). The number of resistors in parallel connnection can be controlled using multipliers. 
    Note that parallel and series resistors can be used simultaneously, but the parallel-ness will be applied to all resistors in series. The cell can be used and routed separately should a more complex combination of resistances be required

    Args:
        pdk (MappedPDK): the process design kit to be used
        width (float, optional): the width of each pfet. Defaults to 5.
        length (float, optional): the length of each pfet. Defaults to 1.
        num_series (int, optional): the number of pfets connected in series. Defaults to 1.
        with_substrate_tap (bool, optional): the presence of substrate tap. Defaults to False.
        with_tie (bool, optional): the presence of tie. Defaults to False.
        with_dnwell (bool, optional): the presence of dnwell. Defaults to False.
        rmult (Optional[int], optional): the routing multiplier (controls routing width). Defaults to None.
        multipliers (int, optional): the number of pfets connected in parallel. Defaults to 1.
        substrate_tap_layers (tuple[str, str], optional): the layers in the substrate tapring. Defaults to ("met2", "met1").
        tie_layers (tuple[str, str], optional): the layers in the tie layer tapring. Defaults to ("met2", "met1").

    Returns:
        Component: an instance of the resistor cell
    """
    toplvl = Component()
    max_sep = pdk.util_max_metal_seperation()
    if num_series == 1:
        pfet_reference = toplvl << pmos(pdk, width=width, length=length, with_substrate_tap=with_substrate_tap, with_tie=with_tie, dnwell=with_dnwell, rmult=rmult, multipliers=multipliers, substrate_tap_layers=substrate_tap_layers, tie_layers=tie_layers, with_dummy=False)
        toplvl.add_ports(pfet_reference.ports, prefix='pfet_')
        
        # short gate and drain 
        diode_connect = toplvl << c_route(pdk, pfet_reference.ports['multiplier_0_gate_W'], pfet_reference.ports['multiplier_0_drain_W'])
        
    else:
        pfet_references = []
        diode_connect_references = []
        pfet_reference_0 = toplvl << pmos(pdk, width=width, length=length, with_substrate_tap=False, with_tie=False, dnwell=False, rmult=rmult, multipliers=multipliers, substrate_tap_layers=substrate_tap_layers, tie_layers=tie_layers, with_dummy=False)
        diode_connect_0 = toplvl << c_route(pdk, pfet_reference_0.ports['multiplier_0_gate_W'], pfet_reference_0.ports['multiplier_0_drain_W'])
        diode_connect_references.append(diode_connect_0)
        
        toplvl.add_ports(pfet_reference_0.ports, prefix='pfet_0_')
        pfet_references.append(pfet_reference_0)
        for i in range(1, num_series):
            pfet_reference = (toplvl << pmos(pdk, width=width, length=length, with_substrate_tap=False, with_tie=False, dnwell=False, rmult=rmult,multipliers=multipliers, substrate_tap_layers=substrate_tap_layers, tie_layers=tie_layers, with_dummy=False)).movey(i * (evaluate_bbox(pfet_reference_0)[1] + max_sep))
            
            pfet_references.append(pfet_reference)
            if i < num_series - 1:
                toplvl.add_ports(pfet_reference.ports, prefix=f'pfet_{i}_')
            
            # short gate and drain 
            diode_connect = toplvl << c_route(pdk, pfet_reference.ports['multiplier_0_gate_W'], pfet_reference.ports['multiplier_0_drain_W'])
            diode_connect_references.append(diode_connect)
            # connect drain and source of previous and present pfet
            if multipliers > 1:
                extension = 1 * ((i % 2) + 1)
            else:
                extension = 0.5 * ((i % 2))
            toplvl << c_route(pdk, pfet_references[i-1].ports['multiplier_0_source_E'], pfet_reference.ports['multiplier_0_drain_E'], extension=extension)
        
        # add tie if tie
        if with_tie:
            tap_separation = max(
                pdk.get_grule("met2")["min_separation"],
                pdk.get_grule("met1")["min_separation"],
                pdk.get_grule("active_diff", "active_tap")["min_separation"],
            )
            tap_separation += pdk.get_grule("n+s/d", "active_tap")["min_enclosure"]
            tap_encloses = (
                (evaluate_bbox(toplvl)[0] + max_sep),
                (evaluate_bbox(toplvl)[1] + max_sep),
            )
            ringtoadd = tapring(
                pdk,
                enclosed_rectangle=tap_encloses,
                sdlayer="n+s/d",
                horizontal_glayer=tie_layers[0],
                vertical_glayer=tie_layers[1],
            )
            tapring_ref = (toplvl << ringtoadd).movey(((evaluate_bbox(pfet_reference_0)[1] + max_sep) * ((num_series - 1)/2) ))
            toplvl.add_ports(tapring_ref.get_ports_list(),prefix="tie_")
            for row in range(multipliers):
                for dummyside, tieside in [("L","W"),("R","E")]:
                    try:
                        toplvl << straight_route(pdk, toplvl.ports[f"multiplier_{row}_dummy_{dummyside}_gsdcon_top_met_W"], toplvl.ports[f"tie_{tieside}_top_met_{tieside}"],glayer2="met1")
                    except KeyError:
                        pass
        
        # add nwell
        nwell_glayer = "dnwell" if with_dnwell else "nwell"
        toplvl.add_padding(
            layers=(pdk.get_glayer(nwell_glayer),),
            default=pdk.get_grule("active_tap", nwell_glayer)["min_enclosure"],
        )
        toplvl = add_ports_perimeter(toplvl, layer=pdk.get_glayer(nwell_glayer),prefix="well_")
        
        # add substrate tap if needed
        if with_substrate_tap:
            substrate_tap_separation = pdk.get_grule("dnwell", "active_tap")[
                "min_separation"
            ]
            substrate_tap_encloses = (
                (evaluate_bbox(toplvl)[0] + max_sep),
                (evaluate_bbox(toplvl)[1] + max_sep),
            )
            ringtoadd = tapring(
                pdk,
                enclosed_rectangle=substrate_tap_encloses,
                sdlayer="p+s/d",
                horizontal_glayer=substrate_tap_layers[0],
                vertical_glayer=substrate_tap_layers[1],
            )
            tapring_ref = (toplvl << ringtoadd).movey(((evaluate_bbox(pfet_reference_0)[1] + max_sep) * ((num_series - 1)/2) ))
            toplvl.add_ports(tapring_ref.get_ports_list(),prefix="guardring_")
            
        toplvl.add_ports(pfet_references[0].get_ports_list(), prefix='port1_')
        toplvl.add_ports(pfet_references[-1].get_ports_list(), prefix='port2_')
        
    return toplvl


# Utilities included in this python file can be used to 
# 1- verify the test_cases
# 2- augment test_cases to synthetically increase the number of data points


import os
import traceback
from pathlib import Path
from typing import Union

from glayout.flow.pdk.mappedpdk import MappedPDK
from glayout.flow.pdk.sky130_mapped import sky130_mapped_pdk
import glayout.syntaxer.dynamic_load


def instantiate_convo(pdk: MappedPDK, convo_file: Union[str, Path]) -> bool:
    """Instantiates a layout for the given conversation file
    
    Args:
        convo_file (str or Path): path to Glayout .convo file

    Returns:
        bool: True if the cell was instantiated
    """
    try:
        # convert NLP to code and pass to show_glayout_code_cell
        session_code = glayout.syntaxer.dynamic_load.run_session(load_conversation=convo_file, restore_and_exit=True)
        comp = glayout.syntaxer.dynamic_load.run_glayout_code_cell(pdk, session_code)
        comp.show()
        # pdk.magic_drc(comp)
        # pdk.lvs_netgen(comp)
        return True
    except Exception as e:
        print(f"Error running session with {convo_file}: {e}")
        print(traceback.format_exc())
    return False


def run_all_tests(test_cases_dir: Union[str, Path] = "./syntax_data/convos"):
    """Run all test cases found in the 'test_cases' directory."""
    # Directory containing test cases
    test_cases_dir = str(Path(test_cases_dir).resolve())
    # Get all files in the directory that end with ".convo"
    convo_files = [f for f in os.listdir(test_cases_dir) if f.endswith(".convo")]
    # run and verify all convos
    for convo_file in convo_files:
        convo_file_path = os.path.join(test_cases_dir, convo_file)
        print(f"Running session with {convo_file_path}")
        instantiate_convo(pdk=sky130_mapped_pdk, convo_file=convo_file_path)


if __name__ == "__main__":
    run_all_tests()

